"""XML format support using the XML Metadata Interchange (XMI) standard."""
import shutil
import uuid as _uuid
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

from lxml import etree

from ragraph.edge import Edge
from ragraph.generic import Annotations, Metadata
from ragraph.graph import Graph
from ragraph.node import Node

here = Path(__file__).parent
schema_doc = etree.parse(str(here / "ragraph.xsd"))
schema = etree.XMLSchema(schema_doc)
schema_root = schema_doc.getroot()

nsmap = schema_root.nsmap
q_id = etree.QName(nsmap["xmi"], "id")
q_graph = etree.QName(nsmap["rc"], "Graph")
q_node = etree.QName(nsmap["rc"], "Node")
q_edge = etree.QName(nsmap["rc"], "Edge")
q_weight = etree.QName(nsmap["rc"], "Weight")
q_annotation = etree.QName(nsmap["rc"], "Annotation")


def _get_xmi_id(obj: Metadata, prefix: str) -> str:
    """Convert a string to a valid XMI id."""
    return prefix + str(obj.uuid)


def _metadata_to_subelements(elem: etree.Element, obj: Metadata):
    """Fetch metadata from object and append to element."""
    name = etree.SubElement(elem, "name")
    name.text = obj.name

    uuid = etree.SubElement(elem, "uuid")
    uuid.text = str(obj.uuid)

    kind = etree.SubElement(elem, "kind")
    kind.text = obj.kind

    for lab in obj.labels:
        label = etree.SubElement(elem, "label")
        label.text = lab

    for k, v in obj.weights.items():
        weight = etree.SubElement(elem, q_weight, attrib=dict(value=str(v)))
        weight.text = k

    elem.extend(_annotations_to_elements(obj.annotations))


def _annotations_to_elements(annotations: Annotations) -> List[etree.Element]:
    """Convert Annotations to XML elements."""
    elems = []
    to_str = {int: "integer", float: "float", bool: "boolean"}

    for k, v in annotations.items():
        elem = etree.Element(q_annotation)

        name_elem = etree.SubElement(elem, "name")
        type_elem = etree.SubElement(elem, "type")
        value_elem = etree.SubElement(elem, "value")

        name_elem.text = k

        if v is None:
            type_elem.text = "none"
        elif type(v) == str:
            type_elem.text = "string"
            value_elem.text = v
        elif type(v) in to_str:
            type_elem.text = to_str[type(v)]
            value_elem.text = str(v).lower()
        else:
            raise TypeError(
                f"Cannot export annotation '{k}'. ",
                "Only annotations whose type is in {str, bool, int, float, None} are ",
                "supported.",
            )

        elems.append(elem)

    return elems


def _node_to_element(node: Node) -> etree.Element:
    """Convert a Node into an XML element."""

    node_id = _get_xmi_id(node, "_n_")
    elem = etree.Element(q_node, {q_id: node_id})

    _metadata_to_subelements(elem, node)

    if node.parent is not None:
        parent = etree.SubElement(elem, "parent")
        parent.text = _get_xmi_id(node.parent, "_n_")

    for c in node.children:
        child = etree.SubElement(elem, "child")
        child.text = _get_xmi_id(c, "_n_")

    if node.is_bus:
        is_bus = etree.SubElement(elem, "is_bus")
        is_bus.text = "true"

    return elem


def _edge_to_element(edge: Edge) -> etree.Element:
    """Convert an Edge to an XML element."""
    edge_id = _get_xmi_id(edge, "_e_")
    elem = etree.Element(q_edge, {q_id: edge_id})

    _metadata_to_subelements(elem, edge)

    source = etree.SubElement(elem, "source")
    source.text = _get_xmi_id(edge.source, "_n_")

    target = etree.SubElement(elem, "target")
    target.text = _get_xmi_id(edge.target, "_n_")

    return elem


def to_xml(
    graph: Graph,
    path: Optional[Union[str, Path]] = None,
    elem: Optional[etree.Element] = None,
    tostring_args: Optional[Dict[str, Any]] = None,
    bundle_schemas: bool = False,
) -> str:
    """Encode Graph to an XML element.

    Arguments:
        graph: Graph to convert to XML.
        path: Optional file path to write XML to.
        elem: Optional object to append the Graph to.
        tostring_args: Optional argument overrides to
            [`lxml.etree.tostring`][lxml.etree.tostring].
        bundle_schemas: When exporting to a file, bundle the .xsd schemas.

    Returns:
        XML element.
    """

    if elem is None:
        parser = etree.XMLParser(remove_blank_text=True)
        doc = etree.parse(str(here / "bare.xml"), parser)
        elem = doc.getroot()

    graph_elem = etree.SubElement(elem, q_graph, nsmap=nsmap)
    _metadata_to_subelements(graph_elem, graph)
    graph_elem.extend(_node_to_element(n) for n in graph.nodes)
    graph_elem.extend(_edge_to_element(e) for e in graph.edges)

    if path is not None:
        p = Path(path)

        args = dict(encoding="UTF-8", xml_declaration=True, pretty_print=True)
        if tostring_args is not None:
            args.update(tostring_args)

        string = etree.tostring(elem, **args)
        p.write_bytes(string)

        if bundle_schemas:
            shutil.copy2(here / "ragraph.xsd", p.parent)
            shutil.copy2(here / "XMI-Canonical.xsd", p.parent)

    return elem


def _elements_to_annotations(elems: Iterable[etree.Element]) -> Annotations:
    """Convert XML elements of type rc:Annotation to an Annotations object."""
    to_value: Dict[str, Callable] = {
        "integer": int,
        "float": float,
        "boolean": bool,
        "none": lambda x: None,
    }

    annots = Annotations()
    for e in elems:
        name = e.find(".//name").text
        cast = to_value.get(e.find(".//type").text, str)
        value = cast(e.find(".//value").text)
        setattr(annots, name, value)

    return annots


def _elements_to_weights(elems: Iterable[etree.Element]) -> Dict[str, Any]:
    """Convert XML elements of rc:Weight type to a weights dictionary."""
    weights = {
        e.text: float(e.attrib["value"]) if "." in e.attrib["value"] else int(e.attrib["value"])
        for e in elems
    }
    return weights


def _element_to_metadata(elem: etree.Element, obj: Metadata):
    """Extract metadata from element."""
    obj.name = elem.find("name").text
    obj.uuid = _uuid.UUID(elem.find("uuid").text)
    obj.kind = elem.find("kind").text
    obj.labels = [i.text for i in elem.findall("label")]
    obj.weights = _elements_to_weights(elem.findall(f"{q_weight}"))
    obj.annotations = _elements_to_annotations(elem.findall(f"{q_annotation}"))


def _element_to_node(elem: etree.Element) -> Node:
    """Convert an XML element to a Node (parent and children stored as strings)."""
    node = Node()
    _element_to_metadata(elem, node)

    parent = elem.find("parent")
    if parent is not None:
        node._parent = parent.text

    node._children = [i.text for i in elem.findall("child")]

    node.is_bus = (
        elem.find("is_bus").text.lower() == "true" if elem.find("is_bus") is not None else False
    )

    return node


def _element_to_edge(elem: etree.Element, node_dict: Dict[str, Any]) -> Edge:
    """Convert an XML element to an Edge object."""
    source = node_dict[elem.find("source").text]
    target = node_dict[elem.find("target").text]
    edge = Edge(source, target)
    _element_to_metadata(elem, edge)
    return edge


def _resolve_node_references(node_dict: Dict[str, Node]):
    """Resolve parent/child nodal string references from names to nodes."""
    for k, v in node_dict.items():
        if isinstance(v._parent, str):
            v._parent = node_dict[v._parent]
        if v._children:
            if isinstance(v._children[0], str):
                v._children = [node_dict[c] for c in v._children]  # type: ignore


def from_xml(
    path: Optional[Union[str, Path]] = None,
    enc: Optional[str] = None,
    elem: Optional[etree.Element] = None,
    validate: bool = True,
) -> Graph:
    """Decode XML file, string, or element into a Graph.

    Arguments:
        path: XML file path.
        enc: XML encoded string.
        elem: XML Element.
        validate: Whether to validate the XML input.

    Returns:
        Graph object.

    Note:
        You should only provide one of `path`, `enc`, or `elem`, which are handled in that order of
        precedence.
    """
    # Parse input XML.
    if path is not None:
        elem = etree.parse(str(Path(path))).getroot()
    elif enc is not None:
        elem = etree.fromstring(enc)
    elif elem is not None:
        assert isinstance(elem, etree._Element)
    else:
        raise ValueError("At least one of the 'path', 'enc', or 'elem' arguments should be set.")

    # Validate if required.
    if validate and not schema.validate(elem):
        raise ValueError(f"Schema validation failed: \n\n{schema.error_log}")

    # Find the Graph XML element.
    g = elem.find(f".//{q_graph}")

    # Parse nodes.
    node_dict = dict()
    for elem in g.findall(f".//{q_node}"):
        node = _element_to_node(elem)
        node_dict[elem.attrib[q_id]] = node
    _resolve_node_references(node_dict)

    # Parse edges.
    edges = [_element_to_edge(e, node_dict) for e in g.findall(f".//{q_edge}")]

    # Build graph and return.
    graph = Graph(nodes=node_dict.values(), edges=edges)
    _element_to_metadata(g, graph)
    return graph



"""# Canopy by Ratio CASE format support"""

import json
import re
from pathlib import Path
from typing import Any, Dict, Generator, List, Optional, Union
from uuid import UUID

from ragraph.generic import Metadata
from ragraph.graph import Edge, Graph, Node

SCHEMA_PAT = re.compile(r"https://canopy.ratio-case.nl/schemas/v1/(\w+).schema.json")


def _match_schema(pat: str) -> Optional[str]:
    match = SCHEMA_PAT.match(pat)
    if match:
        return match.groups()[0]
    return None


def from_canopy(
    path: Union[str, Path],
) -> List[Graph]:
    """Get the graph(s) from a Canopy data file.

    Arguments:
        path: Path of the file to load.

    Returns:
        Graph objects contained in Canopy data formats being either a graph, tab,
        session, or workspace export of Canopy.

    Raises:
        InconsistencyError: if graph is inconsistent.
    """
    data = json.loads(Path(path).read_text())
    schema = _match_schema(data.get("$schema", ""))
    if schema == "workspace":
        graphs = list(_decode_workspace(data))
    elif schema == "session":
        graphs = list(_decode_session(data))
    elif schema == "tab":
        graphs = list(_decode_tab(data))
    else:
        graphs = [_decode_graph(data)]

    for graph in graphs:
        graph.check_consistency(raise_error=True)

    return graphs


def to_canopy(
    graph: Graph,
    path: Optional[Union[Path, str]] = None,
    fmt: str = "session",
) -> Optional[str]:
    """Save graph as a Canopy dataset.

    Arguments:
        graph: Graph to save.
        path: Path to write to.
        fmt: One of 'session' or 'graph'.

    Returns:
        JSON encoded string if no path was provided to write to.
    """
    if fmt == "session":
        data = _encode_session(graph, schema=True)
    else:
        data = _encode_graph(graph, schema=True)
    if path is not None:
        enc = json.dumps(data)
        Path(path).write_text(enc, encoding="utf-8")
        return None
    else:
        enc = json.dumps(data, indent=2)
        return enc


def _encode_metadata(obj: Metadata, schema: bool = True) -> Dict[str, Any]:
    data: Dict[str, Any] = (
        {"$schema": "https://canopy.ratio-case.nl/schemas/v1/metadata.schema.json"}
        if schema
        else dict()
    )
    data.update(
        id=str(obj.uuid),
        name=obj.name,
        kind=obj.kind,
    )
    if obj.labels:
        data.update(labels=list(dict.fromkeys(obj.labels)))
    if obj.weights:
        data.update(weights={str(k): float(v) for k, v in obj.weights.items()})
    adict = obj.annotations.as_dict()
    if adict:
        data.update(annotations=adict)
    return data


def _encode_node(obj: Node, schema: bool = True) -> Dict[str, Any]:
    data: Dict[str, Any] = (
        {"$schema": "https://canopy.ratio-case.nl/schemas/v1/node.schema.json"}
        if schema
        else dict()
    )
    data.update(_encode_metadata(obj, schema=False))
    if obj.is_bus:
        data.update(isBus=bool(obj.is_bus))
    if obj.children:
        data.update(children=[str(n.uuid) for n in obj.children])
    return data


def _encode_edge(obj: Edge, schema: bool = True) -> Dict[str, Any]:
    data: Dict[str, Any] = (
        {"$schema": "https://canopy.ratio-case.nl/schemas/v1/edge.schema.json"}
        if schema
        else dict()
    )
    data.update(_encode_metadata(obj, schema=False))
    data.update(source=str(obj.source.uuid), target=str(obj.target.uuid))
    return data


def _encode_graph(obj: Graph, schema: bool = True) -> Dict[str, Any]:
    data: Dict[str, Any] = (
        {"$schema": "https://canopy.ratio-case.nl/schemas/v1/graph.schema.json"}
        if schema
        else dict()
    )
    data.update(_encode_metadata(obj, schema=False))
    if obj.nodes:
        data.update(
            dict(
                nodes=[_encode_node(n, schema=False) for n in obj.nodes],
            )
        )
    if obj.edges:
        data.update(
            dict(
                edges=[_encode_edge(e, schema=False) for e in obj.edges],
            )
        )
    return data


def _encode_session(graph: Graph, schema: bool = True) -> Dict[str, Any]:
    data: Dict[str, Any] = (
        {"$schema": "https://canopy.ratio-case.nl/schemas/v1/session.schema.json"}
        if schema
        else dict()
    )
    data.update(graph=_encode_graph(graph, schema=False))
    return data


def _decode_node(data: Dict[str, Any]) -> Node:
    opts = data.copy()
    opts.pop("$schema", None)
    opts["uuid"] = opts.pop("id", None)
    opts.pop("isBus", None)
    opts.pop("children", None)
    opts.pop("parent", None)
    return Node(**opts)


def _decode_hierarchy(data: Dict[str, Any], graph: Graph) -> None:
    for node_data in data.get("nodes", []):
        children = node_data.get("children")
        if not children:
            continue
        node = graph.node_uuid_dict[UUID(node_data.get("id"))]
        node.children = [graph.node_uuid_dict[UUID(uuid)] for uuid in children]
    for node_data in data.get("nodes", []):
        if node_data.get("isBus"):
            node = graph.node_uuid_dict[UUID(node_data["id"])]
            if node.parent:
                node.is_bus = True


def _decode_edge(data: Dict[str, Any], graph: Graph) -> Edge:
    opts = data.copy()
    opts.pop("$schema", None)
    opts["uuid"] = opts.pop("id", None)
    opts["source"] = graph.node_uuid_dict[UUID(data["source"])]
    opts["target"] = graph.node_uuid_dict[UUID(data["target"])]
    return Edge(**opts)


def _decode_graph(data: Dict[str, Any]) -> Graph:
    opts = data.copy()
    opts.pop("$schema", None)
    opts["uuid"] = opts.pop("id", None)
    opts.pop("nodes", None)
    opts.pop("edges", None)
    graph = Graph(**opts)
    graph.nodes = [_decode_node(node_data) for node_data in data.get("nodes", [])]
    _decode_hierarchy(data, graph)
    graph.edges = [_decode_edge(edge_data, graph) for edge_data in data.get("edges", [])]
    return graph


def _decode_tab(data: Dict[str, Any]) -> Generator[Graph, None, None]:
    if isinstance(data.get("graph"), dict):
        yield _decode_graph(data["graph"])


def _decode_session(data: Dict[str, Any]) -> Generator[Graph, None, None]:
    if isinstance(data.get("graph"), dict):
        yield _decode_graph(data["graph"])
    for tab_data in data.get("tabs", []):
        if isinstance(tab_data, dict):
            yield from _decode_tab(tab_data)


def _decode_workspace(data: Dict[str, Any]) -> Generator[Graph, None, None]:
    for graph_data in data.get("graphs", []):
        yield _decode_graph(graph_data)
    for session_data in data.get("sessions", []):
        yield from _decode_session(session_data)
    for tab_data in data.get("tabs", []):
        yield from _decode_tab(tab_data)



"""# Elephant Specification Language format support"""

from pathlib import Path
from typing import Union

from raesl.compile import to_graph

from ragraph.graph import Graph


def from_esl(
    *paths: Union[str, Path],
) -> Graph:
    """Convert ESL file(s) into a :obj:`ragraph.graph.Graph`.

    Arguments:
        paths: Paths to resolve into ESL files. May be any number of files and
            directories to scan.

    Returns:
        Instantiated graph.
    """
    return to_graph(*paths, output=None, force=False)



"""# Similarity analysis"""
from typing import Any, Callable, List, Tuple

import numpy as np

from ragraph.analysis import cluster, similarity
from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node


class SimilarityAnalysis:
    """Similarity analysis of nodes based upon mutual mapping relations.

    Arguments:
        cols: List of column nodes.
        rows: List of row nodes.
        edges: List of edges from column nodes to row nodes to be used in similarity
            analysis.
        col_sim_threshold: Column similarity threshold. Values below this threshold are
            pruned from the similarity matrix and the corresponding edges are removed.
            Defaults to 0.0 (no threshold).
        row_sim_threshold: Column similarity threshold. Values below this threshold are
            pruned from the similarity matrix and the corresponding edges are removed.
            Defaults to 0.0 (no threshold).

    Class Attributes:
        similarity_kind: Edge kind for similarity edges. Defaults to ``similarity``.

    Note:
        A mapping matrix relating M column nodes to N row nodes is used as input for
        the similarity analysis.
    """

    similarity_kind = "similarity"

    def __init__(
        self,
        rows: List[Node],
        cols: List[Node],
        edges: List[Edge],
        col_sim_threshold: float = 0.0,
        row_sim_threshold: float = 0.0,
    ):
        self._cols: List[Node] = []
        self._rows: List[Node] = []
        self._edges: List[Edge] = []
        self._col_sim_threshold = 0.0
        self._row_sim_threshold = 0.0
        self._graph: Graph = None  # type: ignore

        self.cols = cols
        self.rows = rows
        self.edges = edges
        self.col_sim_threshold = col_sim_threshold
        self.row_sim_threshold = row_sim_threshold

    @property
    def rows(self) -> List[Node]:
        """List of rows nodes."""
        return self._rows

    @rows.setter
    def rows(self, value: List[Node]):
        kinds = set([n.kind for n in value])
        if len(kinds) > 1:
            raise ValueError(
                f"""All row nodes must be of the same kind. Found node kinds:
                 {kinds}."""
            )

        if self._rows:
            self._rows = value
            self.update_graph()
        else:
            self._rows = value

    @property
    def cols(self) -> List[Node]:
        """List of column nodes."""
        return self._cols

    @cols.setter
    def cols(self, value: List[Node]):
        kinds = set([n.kind for n in value])
        if len(kinds) > 1:
            raise ValueError(
                f"""All columns nodes must be of the same kind. Found node kinds:
                 {kinds}."""
            )

        if self._cols:
            self._cols = value
            self.update_graph()
        else:
            self._cols = value

    @property
    def edges(self) -> List[Edge]:
        """List of edges."""
        return self._edges

    @edges.setter
    def edges(self, value: List[Edge]):
        if self._edges:
            self._edges = value
            self.update_graph()
        else:
            self._edges = value

    @property
    def row_sim_threshold(self) -> float:
        """Similarity threshold. Values below this threshold are pruned from the
        row similarity matrix and the corresponding edges are removed.
        """
        return self._row_sim_threshold

    @row_sim_threshold.setter
    def row_sim_threshold(self, value: float):
        if value < 0.0 or value >= 1.0:
            raise ValueError(
                f"""Similarity threshold must have a value between 0.0 and 1.0. Found
                value: {value}."""
            )

        if self._row_sim_threshold is not None:
            self._row_sim_threshold = float(value)
            self.update_row_similarity()
        else:
            self._row_sim_threshold = float(value)

    @property
    def col_sim_threshold(self) -> float:
        """Similarity threshold. Values below this threshold are pruned from the
        column similarity matrix and the corresponding edges are removed.
        """
        return self._col_sim_threshold

    @col_sim_threshold.setter
    def col_sim_threshold(self, value: float):
        if value < 0.0 or value >= 1.0:
            raise ValueError(
                f"""Similarity threshold must have a value between 0.0 and 1.0. Found
                value: {value}."""
            )

        if self._col_sim_threshold is not None:
            self._col_sim_threshold = float(value)
            self.update_col_similarity()
        else:
            self._col_sim_threshold = float(value)

    @property
    def graph(self) -> Graph:
        """Graph containing similarity edges."""
        if self._graph:
            return self._graph
        else:
            self.update_graph()
            return self._graph

    @property
    def row_similarity_matrix(self) -> np.ndarray:
        """The row similarity matrix based on their mapping row."""
        jaccard = similarity.jaccard_matrix(self.rows, self.row_mapping)
        jaccard[jaccard < self.row_sim_threshold] = 0.0
        return jaccard

    @property
    def col_similarity_matrix(self) -> np.ndarray:
        """The column similarity matrix based on their mapping column."""
        jaccard = similarity.jaccard_matrix(self.cols, self.col_mapping)
        jaccard[jaccard < self.col_sim_threshold] = 0.0
        return jaccard

    def update_graph(self) -> None:
        """Update Internal similarity graph"""
        self._graph = Graph(nodes=self.cols + self.rows, edges=self.edges)
        self.update_col_similarity()
        self.update_row_similarity()

    def update_row_similarity(self) -> None:
        """Update Jaccard Row Similarity Index edges between (clustered) rows."""
        self._update_similarity(self.rows, self.row_similarity_matrix)

    def update_col_similarity(self) -> None:
        """Update Jaccard Column Similarity Index edges between (clustered) columns."""
        self._update_similarity(self.cols, self.col_similarity_matrix)

    def _update_similarity(self, nodes: List[Node], mat: np.ndarray) -> None:
        """Update Jaccard Similarity Index edges between (clustered) nodes."""
        if not self.graph:
            self.update_graph()

        for e in [
            edge
            for edge in self.graph.edges_between_all(nodes, nodes)
            if edge.kind == self.similarity_kind
        ]:
            self.graph.del_edge(e)

        for row, target in enumerate(nodes):
            for col, source in enumerate(nodes):
                if row == col:
                    continue
                sim = mat[row][col]
                if not sim:
                    continue
                self.graph.add_edge(
                    Edge(
                        source,
                        target,
                        kind=self.similarity_kind,
                        weights=dict(similarity=mat[row][col]),
                    )
                )

    def cluster_rows(
        self,
        algo: Callable[[Graph, Any], Tuple[List[Node]]] = cluster.markov,
        **algo_args: Any,
    ) -> None:
        """Cluster column nodes based on their similarity. Updates Graph in-place.

        Arguments:
            algo: Clustering algorithm. Should take a graph as first argument and cluster it
                in-place. Defaults to [`cluster.markov`][ragraph.analysis.cluster.markov].
            **algo_args: Algorithm arguments. See
                [`cluster.markov`][ragraph.analysis.cluster.markov] for sensible defaults.
        """
        self._cluster(self.rows, algo, **algo_args)

    def cluster_cols(
        self,
        algo: Callable[[Graph, Any], Tuple[List[Node]]] = cluster.markov,
        **algo_args: Any,
    ) -> None:
        """Cluster column nodes based on their similarity. Updates Graph in-place.

        Arguments:
            algo: Clustering algorithm. Should take a graph as first argument and cluster it
                in-place. Defaults to [`cluster.markov`][ragraph.analysis.cluster.markov].
            **algo_args: Algorithm arguments. See
                [`cluster.markov`][ragraph.analysis.cluster.markov] for sensible defaults.
        """
        self._cluster(self.cols, algo, **algo_args)

    def _cluster(
        self,
        leafs: List[Node],
        algo: Callable[[Graph, Any], Tuple[List[Node]]] = cluster.markov,
        **algo_args: Any,
    ) -> None:
        """Cluster column nodes based on their similarity. Updates Graph in-place.

        Arguments:
            leafs: List of row or column nodes to be clustered.
            algo: Clustering algorithm. Should take a graph as first argument and cluster it
                in-place. Defaults to [`cluster.markov`][ragraph.analysis.cluster.markov].
            **algo_args: Algorithm arguments. See
                [`cluster.markov`][ragraph.analysis.cluster.markov] for sensible defaults.
        """
        algo(self.graph, leafs=leafs, **algo_args, inplace=True)  # type: ignore

    def row_mapping(self, row: Node) -> List[bool]:
        """Boolean possession checklist for a row node w.r.t.
        [`self.cols`][ragraph.analysis.similarity.SimilarityAnalysis.cols]."""
        return [self.check_mapping(col, row) for col in self.cols]

    def col_mapping(self, col: Node) -> List[bool]:
        """Boolean possession checklist for a column node w.r.t.
        [`self.rows`][ragraph.analysis.similarity.SimilarityAnalysis.rows]."""
        return [self.check_mapping(col, row) for row in self.rows]

    def check_mapping(self, col: Node, row: Node) -> bool:
        """Check whether a column node maps to a row node."""
        if len(self.graph.directed_edges[col.name][row.name]) > 0:
            return True
        else:
            return False



"""Abstracted analysis base classes and methods."""

import inspect
from abc import ABC, abstractmethod
from copy import deepcopy
from functools import wraps
from logging import DEBUG
from textwrap import dedent
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

from ragraph.analysis import _utils, logger
from ragraph.generic import Bound
from ragraph.graph import Graph
from ragraph.node import Node


class Cast(ABC):
    """Cast for parameters."""

    @abstractmethod
    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def __call__(self, value: Any) -> Any:
        """Cast a value."""
        pass


class GraphCast(Cast):
    """Just check whether the argument is a Graph."""

    def __init__(self):
        pass

    def __call__(self, value: Any) -> Graph:
        if not isinstance(value, Graph):
            raise ValueError(f"Please supply a Graph object, got '{value}'.")
        return value


class NodeCast(Cast):
    """Cast node names into instances.

    Arguments:
        graph: Graph to fetch nodes from.
    """

    def __init__(self, graph: Optional[Graph]):
        self.graph = graph

    def __call__(self, value: Any) -> Optional[Node]:
        if value is None:
            return value
        if self.graph is None:
            raise ValueError("Graph must be set before cast is possible.")
        if isinstance(value, Node):
            if value.name not in self.graph or value not in self.graph.nodes:
                raise ValueError(f"Node '{value}' does not occur in the graph.")
            return value
        return self.graph.node_dict[value]


class NodesCast(Cast):
    """Cast leaf node names into node instances. If an empty list is provided, take
    child nodes below a specific root node or all leafs in the graph.
    """

    def __init__(self, graph: Optional[Graph], root: Optional[Union[Node, str]] = None):
        self.graph = graph
        self.root = root

    def __call__(self, value: Optional[List[str]] = None) -> List[Node]:
        if self.graph is None:
            raise ValueError("Graph must be set before cast is possible.")
        cast = NodeCast(self.graph)
        if value is None:
            root = cast(self.root)
            nodes = self.graph.leafs if root is None else root.children
        else:
            nodes = [n for n in (cast(node) for node in value) if n is not None]
        return nodes


class LeafsCast(Cast):
    """Cast leaf node names into node instances. If an empty list is provided, take
    leaf nodes below a specific root node or all leafs in the graph.
    """

    def __init__(self, graph: Optional[Graph], root: Optional[str] = None):
        self.graph = graph
        self.root = root

    def __call__(self, value: Optional[List[str]] = None) -> List[Node]:
        if self.graph is None:
            raise ValueError("Graph must be set before cast is possible.")
        cast = NodeCast(self.graph)
        if value is None:
            root = cast(self.root)
            leafs = (
                self.graph.leafs if root is None else [n for n in root.descendant_gen if n.is_leaf]
            )
        else:
            leafs = [n for n in (cast(leaf) for leaf in value) if n is not None]
        return leafs


class MethodCast(Cast):
    """Cast a Callable to ensure the method is fetched when it's an Analysis object."""

    def __init__(self):
        pass

    def __call__(self, value: Callable):
        return value.method if isinstance(value, Analysis) else value


class Parameter:
    """Analysis parameter.

    Arguments:
        name: Argument name.
        description: Argument description.
        type: Argument instance type.
        default: Default argument value.
        cast: Optional cast to use for incoming execution values.
        enum: Allowed values if this parameter is an enumeration.
        lower: Lower bound for numeric values (inclusive).
        upper: Upper bound for numeric values (inclusive).
    """

    def __init__(
        self,
        name: str,
        type: Any,
        description: str = "",
        default: Any = None,
        cast: Any = None,
        enum: Optional[Set[Any]] = None,
        lower: Optional[Bound] = None,
        upper: Optional[Bound] = None,
    ):
        self.name = name
        self.description = description
        self.type = type
        self.lower = lower
        self.upper = upper

        # Use built-in conversion for trivial types.
        if cast is None and type in (float, int, str, bool):
            self.cast = type
        else:
            self.cast = cast

        # Cast default, enum, and bounds if possible, else just use input.
        if self.cast is None:
            self.default = default
            self.enum = set(enum) if enum is not None else None
        else:
            self.default = self.cast(default) if default is not None else None
            self.enum = {self.cast(i) for i in enum} if enum is not None else None

    def __str__(self) -> str:
        msg = f"{self.name}: {self.description} "

        if self.default is not None:
            msg += f"Defaults to {self.default}. "

        if self.enum:
            msg += f"Must be one of: '{self.enum}'. "

        if self.lower is not None or self.upper is not None:
            lb = "{}{}".format(
                "(" if self.lower is None or self.lower.inclusive else "[",
                "-inf" if self.lower is None else self.lower.value,
            )
            ub = "{}{}".format(
                "inf" if self.upper is None else self.upper.value,
                ")" if self.upper is None or not self.upper.inclusive else "]",
            )
            msg += f"Bounds: {lb}, {ub}. "

        msg += f"Type should satisfy {self.type}. ".replace("typing.", "")
        return msg

    def parse(self, value: Any) -> Any:
        """Check and parse a value."""
        if self.cast is not None:
            try:
                value = self.cast(value)
            except Exception:
                raise ValueError(
                    f"Error while casting {self.name} value '{value}' using "
                    f"'{self.cast}' into a '{self.type}'."
                )
        if self.enum is not None and value not in self.enum:
            raise ValueError(f"Value '{value}' is not allowed, try one of '{self.enum}'.")
        if self.lower is not None:
            assert value > self.lower
        if self.upper is not None:
            assert value < self.upper
        return value


graph = Parameter(
    "graph",
    Graph,
    description="Graph data object.",
    cast=GraphCast(),
)

root = Parameter(
    "root",
    Optional[Union[str, Node]],
    description="Root node to fetch nodes from.",
    cast=NodeCast(None),
)

nodes = Parameter(
    "nodes",
    Optional[Union[List[str], List[Node]]],
    description="Selected nodes to apply analysis on.",
    cast=NodesCast(None, None),
)


leafs = Parameter(
    "leafs",
    Optional[Union[List[str], List[Node]]],
    description="Selected leaf nodes.",
    cast=LeafsCast(None, None),
)


edge_weights = Parameter(
    "edge_weights",
    Optional[List[str]],
    description="Edge weights to consider during calculations.",
)


names = Parameter(
    "names",
    bool,
    description="Whether to return node names instead of instances.",
    default=False,
)


inplace = Parameter(
    "inplace",
    bool,
    description="Whether to update the input graph inplace or return a copy.",
    default=True,
)

inherit = Parameter(
    "inherit",
    bool,
    description="Whether edges between descendants of nodes should be taken "
    + "into account (if applicable).",
    default=True,
)

loops = Parameter(
    "loops",
    bool,
    description="Whether self-loop edges should be taken into account (if applicable).",
    default=False,
)

safe = Parameter(
    "safe",
    bool,
    description="Whether to check and cast input arguments to their appropriate types.",
    default=True,
)


class Analysis:
    """Analysis interface.

    Arguments:
        name: Analysis name.
        description: Analysis description.
        parameters: Analysis parameter mapping.
    """

    _default_parameters: Dict[str, Parameter] = dict()
    _returns: Optional[List[str]] = None

    def __init__(
        self,
        name: str,
        description: str = "",
        parameters: Optional[Dict[str, Parameter]] = None,
    ):
        self.name = name
        self.description = dedent(description)

        # Store parameters as a dictionary mapping
        self.parameters = deepcopy(self._default_parameters)
        if parameters is not None:
            self.parameters.update(deepcopy(parameters))

        # Initialize method with a placeholder.
        self.method = lambda *a, **kw: None

    def __str__(self) -> str:
        # Basic name and description
        msg = f"{self.name}"

        if self.description:
            msg += f"\n\n{self.description}"

        # Arguments section
        msg += "\n\nArguments:\n    "
        msg += "\n    ".join([str(p) for p in self.parameters.values()])

        # Returns section
        if self._returns is not None:
            msg += "\n\nReturns:\n    "
            msg += "\n    ".join([str(p) for p in self._returns])

        # Final newline
        msg += "\n"

        # Return with some precaution cleanup
        return msg.replace("\n\n\n", "\n\n")

    def __repr__(self) -> str:
        return str(self)

    def __call__(self, func: Callable):
        """Wrap an analysis function to apply docstring and parameter parsing."""
        self._parse_signature(func)
        self._apply_docstring(func)
        self.method = self.wrap(func)
        self.method.__dict__["analysis"] = self
        return self.method

    def _parse_signature(self, func: Callable):
        """Check whether the signature of the function matches with the parameters."""
        sig = inspect.signature(func)

        ref_keys = set(self.parameters.keys())

        diff = ref_keys.symmetric_difference(sig.parameters.keys()) - {"kwargs"}
        if diff:
            raise ValueError(
                f"Parameter mismatch. Analysis has '{ref_keys}', "
                + f"function has '{sig.parameters.keys()}'. Difference: '{diff}'."
            )

        for pname, param in self.parameters.items():
            try:
                f_param = sig.parameters[pname]
            except KeyError:
                raise ValueError(f"Parameter {pname} is missing in the signature of '{func}'.")

            # Inherit default kwarg values.
            func.__annotations__[pname] = param.type
            if f_param.default != inspect._empty:
                param.default = f_param.default

    def _apply_docstring(self, func: Callable):
        """Apply new docstring."""
        func.__doc__ = str(self)

    def _parse_parameters(self, kw: Dict[str, Any]):
        """Parse given arguments using the Parameter specifications."""
        params = self.parameters

        for pname, param in params.items():
            if pname not in kw:
                kw[pname] = param.default

        if "graph" in params:
            kw["graph"] = params["graph"].cast(kw.get("graph"))

        if "root" in params:
            params["root"].cast.graph = kw.get("graph")
            kw["root"] = params["root"].cast(kw.get("root"))

        if "nodes" in params:
            params["nodes"].cast.graph = kw.get("graph")
            params["nodes"].cast.root = kw.get("root")

        if "leafs" in params:
            params["leafs"].cast.graph = kw.get("graph")
            params["leafs"].cast.root = kw.get("root")

        for pname, param in params.items():
            kw[pname] = param.parse(kw.get(pname))

    @abstractmethod
    def wrap(self, func: Callable) -> Callable:
        """Wrapper function to wrap analysis method with."""
        pass

    def log(self, msg: str, level: int = DEBUG):
        """Dispatch a logging message."""
        logger.log(level, f"{self.name}: \n{msg}\n")


class BusAnalysis(Analysis):
    _default_parameters = {
        p.name: p for p in [graph, root, leafs, inherit, loops, edge_weights, names, safe]
    }
    _returns = [
        "Detected bus nodes (or names thereof).",
        "Remaining nonbus nodes (or names thereof).",
    ]

    def wrap(self, func: Callable) -> Callable:
        """Wrapper for all bus detection algorithms.

        Returns:
            Function that preprocesses the arguments and handles some typical/trivial
            cases where nodes are not set or too few to distinguish a bus at all.
        """

        @wraps(func)
        def wrapper(
            *args, **kw
        ) -> Union[Tuple[List[str], List[str]], Tuple[List[Node], List[Node]]]:
            # Process arguments.
            kw.update(zip(func.__code__.co_varnames, args))
            if kw.get("safe", True):
                self._parse_parameters(kw)

            # Trivial case.
            leafs = kw["leafs"]
            if len(leafs) < 3:
                self.log(f"Trivial case, too little nodes {len(leafs)} < 3.")
                bus_leafs, nonbus_leafs = [], leafs
            # Nontrivial case, need to execute:
            else:
                self.log(
                    "Applying bus analysis to leafs " + f"'{[n.name for n in kw['leafs']]}'..."
                )
                bus_leafs, nonbus_leafs = func(**kw)

            # Postprocess results.
            bus_names = [n.name for n in bus_leafs]
            nonbus_names = [n.name for n in nonbus_leafs]
            self.log(f"BUS:    {bus_names}\nNONBUS: {nonbus_names}")

            if kw["names"]:
                return bus_names, nonbus_names
            else:
                return bus_leafs, nonbus_leafs

        return wrapper


class ClusterAnalysis(Analysis):
    _default_parameters = {
        p.name: p
        for p in [
            graph,
            root,
            leafs,
            inherit,
            loops,
            edge_weights,
            inplace,
            names,
            safe,
        ]
    }

    def wrap(self, func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kw) -> Tuple[Graph, Union[List[str], List[Node]]]:
            # Process arguments.
            kw.update(zip(func.__code__.co_varnames, args))
            if kw.get("safe", True):
                self._parse_parameters(kw)

            # Create a deepcopy and update all nodal arguments to the new nodes.
            if not kw["inplace"]:
                kw["graph"] = deepcopy(kw["graph"])
                kw["leafs"] = [kw["graph"].node_dict[n.name] for n in kw["leafs"]]
                kw["root"] = kw["graph"].node_dict[kw["root"].name] if kw.get("root") else None

            # Trivial cases where there is nothing to cluster.
            if len(kw["leafs"]) < 2:
                self.log("Trivial case, less than 2 nodes provided.")
                graph = kw["graph"]
                clusters = kw["leafs"]

            # Non-trivial case, need to execute:
            else:
                # Roots are kept as a list from hereon.
                roots = [kw["root"]] if kw.get("root") else None

                # Cut leafs loose, but preserve root(s) if set.
                for leaf in kw["leafs"]:
                    _utils.unset_parent(kw["graph"], leaf, roots)

                # Execution
                self.log("Applying cluster algorithm...")
                graph, clusters = func(**kw)
                self.log(f"Found {len(clusters)} clusters for {len(kw['leafs'])} leafs.")

            # Postprocess results.
            if kw.get("root"):
                self.log("Setting found clusters as children of root...")
                _utils.set_children(graph, kw["root"], clusters)
                clusters = [kw["root"]]

            cluster_names = [n.name for n in clusters]
            self.log(f"Cluster names: {cluster_names}")
            if kw["names"]:
                return graph, cluster_names
            return graph, clusters

        return wrapper


class SequenceAnalysis(Analysis):
    _default_parameters = {
        p.name: p
        for p in [
            graph,
            root,
            nodes,
            inherit,
            loops,
            edge_weights,
            inplace,
            names,
            safe,
        ]
    }

    def wrap(self, func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kw) -> Tuple[Graph, Union[List[str], List[Node]]]:
            # Process arguments.
            kw.update(zip(func.__code__.co_varnames, args))
            if kw.get("safe", True):
                self._parse_parameters(kw)

            # Preprocessing
            dim = len(kw["nodes"])

            # Trivial case:
            if dim < 2:
                self.log("Trivial case, less than 2 nodes.")
                graph, nodes = kw["graph"], kw["nodes"]
            # Non-trivial case, need to execute:
            else:
                self.log("Applying sequencing algorithm...")
                graph, nodes = func(**kw)

            # Postprocessing
            node_names = [n.name for n in nodes]
            self.log(f"Result: {node_names}.")

            if kw["inplace"] and kw["root"]:
                assert len(nodes) == len(kw["root"].children)
                _utils.set_children(graph, kw["root"], nodes)
                self.log("Re-ordered root node's children in-place.")

            if kw["names"]:
                return graph, node_names
            return graph, nodes

        return wrapper


class BranchsortAnalysis(Analysis):
    _default_parameters = {
        p.name: p
        for p in [
            graph,
            root,
            nodes,
            leafs,
            inherit,
            loops,
            edge_weights,
            Parameter(
                "algo",
                Callable,
                description="Sequencing algorithm to use.",
                cast=MethodCast(),
            ),
            Parameter(
                "algo_args",
                Optional[Dict[str, Any]],
                description="Parameters to pass to the selected sequencing algorithm.",
                cast=lambda value: dict() if value is None else dict(value),
            ),
            Parameter(
                "recurse",
                bool,
                description="Whether to recursively sort branches downwards.",
                default=True,
            ),
            inplace,
            names,
            safe,
        ]
    }

    def wrap(self, func: Callable):
        @wraps(func)
        def wrapper(
            *args, **kw
        ) -> Tuple[Graph, Union[List[str], List[Node]], Union[List[str], List[Node]]]:
            # Process arguments.
            kw.update(zip(func.__code__.co_varnames, args))
            if kw.get("safe", True):
                self._parse_parameters(kw)

            # Preprocessing
            dim = len(kw["nodes"])

            # Trivial case:
            if dim == 0:
                graph, nodes, leafs = kw["graph"], [], []
            # Non-trivial case, need to execute:
            else:
                graph, nodes, leafs = func(**kw)

            # Postprocessing
            node_names = [n.name for n in nodes]
            leaf_names = [n.name for n in leafs]
            self.log(f"Node sequence: {node_names}")
            self.log(f"Leaf sequence: {leaf_names}")

            if kw["inplace"] and kw["root"]:
                assert len(nodes) == len(kw["root"].children)
                _utils.set_children(graph, kw["root"], nodes)
                self.log("Re-ordered root node's children in-place.")

            if kw["names"]:
                return graph, node_names, leaf_names
            return graph, nodes, leafs

        return wrapper



"""# Sigma analysis"""

import enum
from copy import deepcopy
from typing import Iterable

from ragraph.edge import Edge
from ragraph.generic import Metadata
from ragraph.graph import Graph
from ragraph.node import Node


class SigmaMode(enum.StrEnum):
    """Aggregation mode for sigma analysis, absolute counted occurrences or an average per graph."""

    ABSOLUTE = enum.auto()
    AVERAGE = enum.auto()


def sigma_graph(
    graphs: Iterable[Graph], count_prefix: str = "sigma", mode: SigmaMode = SigmaMode.ABSOLUTE
) -> Graph:
    """Get the sigma (summed) graph based on the leaf nodes and the edges in multiple graphs.

    Arguments:
        graphs: Graphs to sum and count edge occurrences in.
        mode: Whether to count absolute occurrence values or an average per graph.
        count_prefix: Which weight key (prefix) to store the occurrence values under. Also used as
            a prefix for label counts that are stored as weights using as well.

    Note:
        Summations are done over (leaf) node names. Edges in an input graph are aggregated first
        using the [`add_meta` method][ragraph.analysis.comparison.add_meta]
        before being aggregated into the resulting graph.
    """
    accumulator = Graph(name="sigma")
    num = 0
    for graph in graphs:
        # Override that first one's results because of default values for a Graph.
        add_graph(accumulator, graph, count_prefix, num == 0)
        num += 1
    if mode == SigmaMode.AVERAGE and num > 0:
        div = 1.0 / num
        for k, v in accumulator.weights.items():
            if k.startswith(f"{count_prefix}_") or k == count_prefix:
                accumulator.weights[k] = div * v
        for n in accumulator.nodes:
            for k, v in n.weights.items():
                if k.startswith(f"{count_prefix}_") or k == count_prefix:
                    n.weights[k] = div * v
        for e in accumulator.edges:
            for k, v in e.weights.items():
                if k.startswith(f"{count_prefix}_") or k == count_prefix:
                    e.weights[k] = div * v
    return accumulator


def add_graph(accumulator: Graph, item: Graph, count_prefix: str, is_first: bool = False) -> Graph:
    """Add a graph to an accumulator graph. Occurrences are counted under the 'count_prefix'."""
    add_meta(accumulator, item, count_prefix, is_first=is_first)
    add_nodes(accumulator, item, count_prefix)
    add_edges(accumulator, item, count_prefix)
    return accumulator


def add_meta(accumulator: Metadata, item: Metadata, count_prefix: str, is_first: bool = False):
    """Add metadata of another item to the accumulator in-place. Annotations are updated and the
    last observed value for any key is kept. Occurrence value is tracked under the 'count_prefix'.
    """
    if is_first:
        initialize_meta(accumulator, item, count_prefix)
    else:
        increment_meta(accumulator, item, count_prefix)


def initialize_meta(accumulator: Metadata, item: Metadata, count_prefix: str):
    """Initialize an accumulator instance with the info from the first item."""
    accumulator.kind = item.kind
    accumulator.labels = deepcopy(item.labels)
    accumulator.weights = deepcopy(item.weights)
    accumulator.weights[count_prefix] = 1
    for label in item.labels:
        k = f"{count_prefix}_label_{label}"
        accumulator.weights[k] = item.weights.get(k, 1)
    accumulator.annotations = deepcopy(item.annotations)


def increment_meta(accumulator: Metadata, item: Metadata, count_prefix: str):
    """Increment the occurrence and label counts as weights in the accumulator with an item."""
    if accumulator.kind != item.kind:
        raise ValueError(f"kinds don't match: '{accumulator.kind}' and '{item.kind}'")

    # Update labels to superset.
    accumulator.labels = sorted(set(accumulator.labels + item.labels))

    # Sum weights
    for k, v in item.weights.items():
        accumulator.weights[k] = accumulator.weights.get(k, 0) + v
    accumulator.annotations.update(item.annotations)

    if count_prefix in item.weights:
        pass  # Is already incremented when summing weights above.
    else:
        accumulator.weights[count_prefix] = accumulator.weights.get(count_prefix, 0) + 1

    # Increment label counts.
    for label in item.labels:
        k = f"{count_prefix}_label_{label}"
        if k in item.weights:
            continue  # Is already summed with weights.
        accumulator.weights[k] = accumulator.weights.get(k, 0) + 1


def add_nodes(accumulator: Graph, item: Graph, count_prefix: str):
    """Add observed nodes in item to the accumulator in-place and keep track of an occurrence count.

    Based on node names.

    Metadata for existing nodes is added using
    [`add_meta` method][ragraph.analysis.comparison.add_meta].
    """
    for n in item.leafs:
        try:
            add_meta(accumulator[n.name], n, count_prefix)
        except KeyError:
            node = Node(name=n.name)
            add_meta(node, n, count_prefix, is_first=True)
            accumulator.add_node(node)


def add_edges(accumulator: Graph, item: Graph, count_prefix: str):
    """Add observed edges in item to the accumulator in-place.

    Based on node names.

    Metadata for existing nodes is added using
    [`add_meta` method][ragraph.analysis.comparison.add_meta].
    """
    for source_name, tgt_map in item.directed_edges.items():
        for target_name, item_edges in tgt_map.items():
            # Take a deepcopy of the list of edges in the item and aggregate those first.
            item_edges = deepcopy(item_edges)
            item_edge = item_edges.pop()
            for e in item_edges:
                add_meta(item_edge, e, count_prefix)

            # Now check whether we already accumulated something here, or add a new entry.
            edges = accumulator.directed_edges[source_name][target_name]
            if len(edges):
                add_meta(edges[0], item_edge, count_prefix)
            else:
                source_node = accumulator.node_dict[source_name]
                target_node = accumulator.node_dict[target_name]
                edge = Edge(source_node, target_node)
                add_meta(edge, item_edge, count_prefix, is_first=True)
                accumulator.add_edge(edge)



"""# RaGraph colors and palettes

This module contains several convenient methods to obtain color palettes as lists of
hex-coded colors. The defaults for categorical and continuous data are
[`get_categorical`][ragraph.colors.get_categorical] and
[`get_continuous`][ragraph.colors.get_continuous], respectively.

Additional methods are included, most of which are based on the
[`cubehelix palette`][ragraph.colors.cubehelix.cubehelix_palette] with certain preset options.
[`get_hue`][ragraph.colors.get_hue] is a great way to get a sequential color palette for any hue
where the hue value corresponds to anything between red (1.0), green (2.0) or blue (3.0).

In our plots, these colors are *not* interpolated. This opens up an easy way to visualize your
colors in a discrete fashion, since you merely need to supply a list with less colors (e.g. the
desired number of discrete steps).
"""
from typing import Any, Dict, List

from ragraph.colors.cubehelix import cubehelix_palette


def get_categorical(n_colors: int = 10, **kwargs) -> List[str]:
    """Get a color palette suitable for categorical data.

    Arguments:
        n_colors: Number of colors to generate.

    Returns:
        A list of colors as hex codes.

    Note:
        Additional keyword arguments are passed on to
        [`cubehelix_palette`][ragraph.colors.cubehelix.cubehelix_palette].
    """
    n_colors = max(n_colors, 1)
    base_light = 0.6
    if n_colors <= 10:
        dark = base_light
        light = base_light
    else:
        excess = n_colors - 10
        dark = max(0.3, base_light - excess * 0.05)
        light = min(0.8, base_light + excess * 0.05)
    opts: Dict[str, Any] = dict(
        n_colors=n_colors,
        start=2.8,
        rotations=(n_colors - 1) / n_colors,
        dark=light,  # Get the darker colors first.
        light=dark,
        gamma=1.0,
        saturation=1.8,
        categorical=True,
    )
    opts.update(kwargs)
    return cubehelix_palette(**opts)


def get_continuous(n_colors: int = 256, **kwargs) -> List[str]:
    """Get a default color palette suitable for continuous valued data.

    Arguments:
        n_colors: Number of colors to generate.

    Returns:
        A list of colors as hex codes.

    Note:
        Additional keyword arguments are passed on to
        [`cubehelix_palette`][ragraph.colors.cubehelix.cubehelix_palette].
    """
    n_colors = max(n_colors, 1)
    opts: Dict[str, Any] = dict(
        n_colors=n_colors,
        start=2.25,
        end=3.45,
        dark=0.2,
        light=0.85,
        gamma=1.0,
        saturation=1.2,
        categorical=False,
    )
    opts.update(kwargs)
    return cubehelix_palette(**opts)


def get_hue(n_colors: int = 256, hue: float = 1.0, **kwargs) -> List[str]:
    """Get a sequential color palette with the given hue.

    Arguments:
        n_colors: Number of colors to generate.
        hue: Hue for this sequential color palette.

    Returns:
        A list of colors as hex codes.

    Note:
        Additional keyword arguments are passed on to
        [`cubehelix_palette`][ragraph.colors.cubehelix.cubehelix_palette].
    """
    n_colors = max(n_colors, 1)
    opts: Dict[str, Any] = dict(
        n_colors=n_colors,
        start=hue,
        end=hue,
        dark=0.25,
        light=1.0,
        gamma=1.0,
        saturation=1.8,
        categorical=False,
    )
    opts.update(**kwargs)
    return cubehelix_palette(**opts)


def get_purple(n_colors: int = 256, **kwargs) -> List[str]:
    """Get a purple sequential color palette.

    Arguments:
        n_colors: Number of colors to generate.

    Returns:
        A list of colors as hex codes.

    Note:
        Additional keyword arguments are passed on to
        [`cubehelix_palette`][ragraph.colors.cubehelix.cubehelix_palette].
    """
    opts: Dict[str, Any] = dict(hue=0.15, n_colors=n_colors)
    opts.update(**kwargs)
    return get_hue(**opts)


def get_red(n_colors: int = 256, **kwargs) -> List[str]:
    """Get a red sequential color palette.

    Arguments:
        n_colors: Number of colors to generate.

    Returns:
        A list of colors as hex codes.

    Note:
        Additional keyword arguments are passed on to
        [`cubehelix_palette`][ragraph.colors.cubehelix.cubehelix_palette].
    """
    opts: Dict[str, Any] = dict(hue=0.95, n_colors=n_colors)
    opts.update(**kwargs)
    return get_hue(**opts)


def get_orange(n_colors: int = 256, **kwargs) -> List[str]:
    """Get an orange sequential color palette.

    Arguments:
        n_colors: Number of colors to generate.

    Returns:
        A list of colors as hex codes.

    Note:
        Additional keyword arguments are passed on to
        [`cubehelix_palette`][ragraph.colors.cubehelix.cubehelix_palette].
    """
    opts: Dict[str, Any] = dict(hue=1.2, n_colors=n_colors)
    opts.update(**kwargs)
    return get_hue(**opts)


def get_green(n_colors: int = 256, **kwargs) -> List[str]:
    """Get a green sequential color palette.

    Arguments:
        n_colors: Number of colors to generate.

    Returns:
        A list of colors as hex codes.

    Note:
        Additional keyword arguments are passed on to
        [`cubehelix_palette`][ragraph.colors.cubehelix.cubehelix_palette].
    """
    opts: Dict[str, Any] = dict(hue=1.8, n_colors=n_colors)
    opts.update(**kwargs)
    return get_hue(**opts)


def get_cyan(n_colors: int = 256, **kwargs) -> List[str]:
    """Get a cyan sequential color palette.

    Arguments:
        n_colors: Number of colors to generate.

    Returns:
        A list of colors as hex codes.

    Note:
        Additional keyword arguments are passed on to
        [`cubehelix_palette`][ragraph.colors.cubehelix.cubehelix_palette].
    """
    opts: Dict[str, Any] = dict(hue=2.45, n_colors=n_colors)
    opts.update(**kwargs)
    return get_hue(**opts)


def get_blue(n_colors: int = 256, **kwargs) -> List[str]:
    """Get a blue sequential color palette.

    Arguments:
        n_colors: Number of colors to generate.

    Returns:
        A list of colors as hex codes.

    Note:
        Additional keyword arguments are passed on to
        [`cubehelix_palette`][ragraph.colors.cubehelix.cubehelix_palette].
    """
    opts: Dict[str, Any] = dict(hue=2.85, n_colors=n_colors)
    opts.update(**kwargs)
    return get_hue(**opts)


def get_diverging_hues(
    n_colors: int = 257,
    lower_hue: float = 0.95,
    upper_hue: float = 2.85,
    midpoint: float = 0.5,
    **kwargs,
) -> List[str]:
    """Get a diverging color palette between two hues.

    Arguments:
        n_colors: Number of colors (will always return an uneven number of colors).
        lower_hue: Lower bound hue value.
        upper_hue: Upper bound hue value.
        midpoint: Fraction between [0.0, 1.0] where to put the midpoint.

    Returns:
        A list of colors as hex codes.

    Note:
        Additional keyword arguments are passed on to
        [`cubehelix_palette`][ragraph.colors.cubehelix.cubehelix_palette].
    """
    if n_colors <= 1:
        return ["#ffffff"]

    midindex = round(midpoint * (n_colors - 1))
    lower = get_hue(hue=lower_hue, n_colors=midindex + 1, **kwargs)
    upper = get_hue(hue=upper_hue, n_colors=n_colors - midindex, **kwargs)
    return lower[::-1] + upper[1:]


def get_diverging_redblue(n_colors: int = 257, **kwargs) -> List[str]:
    """Get a red-white-blue diverging color palette.

    Arguments:
        n_colors: Number of colors to generate.

    Returns:
        A list of colors as hex codes.

    Note:
        Additional keyword arguments are passed on to
        [`cubehelix_palette`][ragraph.colors.cubehelix.cubehelix_palette].
    """
    opts: Dict[str, Any] = dict(lower_hue=0.95, upper_hue=2.85, n_colors=n_colors)
    opts.update(kwargs)
    return get_diverging_hues(**opts)


def get_diverging_orangecyan(n_colors: int = 257, **kwargs) -> List[str]:
    """Get a orange-white-cyan diverging color palette.

    Arguments:
        n_colors: Number of colors to generate.

    Returns:
        A list of colors as hex codes.

    Note:
        Additional keyword arguments are passed on to
        [`cubehelix_palette`][ragraph.colors.cubehelix.cubehelix_palette].
    """
    opts: Dict[str, Any] = dict(lower_hue=1.2, upper_hue=2.45, n_colors=n_colors)
    opts.update(kwargs)
    return get_diverging_hues(**opts)


def get_diverging_purplegreen(n_colors: int = 257, **kwargs) -> List[str]:
    """Get a purple-white-green diverging color palette.

    Arguments:
        n_colors: Number of colors to generate.

    Returns:
        A list of colors as hex codes.

    Note:
        Additional keyword arguments are passed on to
        [`cubehelix_palette`][ragraph.colors.cubehelix.cubehelix_palette].
    """
    opts: Dict[str, Any] = dict(lower_hue=0.15, upper_hue=1.8, n_colors=n_colors)
    opts.update(kwargs)
    return get_diverging_hues(**opts)


def get_colormaps(
    n_colors: int = 1,
    kind: str = "categorical",
    amount: int = 1,
    flip: bool = False,
) -> List[List[str]]:
    """Get multiple colormaps of a certain kind.

    Arguments:
        n_colors: Number of colors per colormap to generate.
        kind: Kind of colormap. One of 'categorical', 'sequential', or 'diverging'.
        amount: Number of colormaps to generate.
        flip: Whether to reverse the generated colormaps.
    """
    if kind == "categorical":
        return [get_categorical(n_colors=n_colors)]
    elif kind == "sequential":
        if amount == 1:
            colormaps = [get_continuous(n_colors=n_colors)]
        else:
            colormaps = [
                get_hue(n_colors=n_colors, hue=2.85 + 3 * i / amount) for i in range(amount)
            ]
    elif kind == "diverging":
        colormaps = [
            get_diverging_hues(
                n_colors=n_colors,
                lower_hue=0.95 - 1.0 * i / amount,
                upper_hue=2.85 - 2 * i / amount,
            )
            for i in range(amount)
        ]
    else:
        allowed = ["categorical", "sequential", "diverging"]
        raise ValueError(f"Unknown colormap kind. Pick one of '{allowed}'.")

    if flip:
        return [colormap[::-1] for colormap in colormaps]
    else:
        return colormaps



"""
# Comparison analysis

Comparison provides classes for comparing [`Graph` objects][ragraph.graph.Graph] to find the
commonalities (sigma) and differences (delta).

## Comparison methods

- [`sigma_graph`][ragraph.analysis.comparison.sigma_graph] to calculate the commonalities (i.e. sum)
  of a set of graphs with comparable node names.
- [`delta_graph`][ragraph.analysis.comparison.delta_graph] to calculate the differences and overlap
  between two graphs with comparable node names.
"""

from ragraph.analysis.comparison._delta import delta_graph
from ragraph.analysis.comparison._sigma import (
    SigmaMode,
    add_edges,
    add_graph,
    add_meta,
    add_nodes,
    increment_meta,
    initialize_meta,
    sigma_graph,
)
from ragraph.analysis.comparison.utils import (
    EdgeDescriptor,
    EdgeDescriptorLike,
    NodeDescriptor,
    NodeDescriptorLike,
    TagMode,
    tag,
)

__all__ = [
    "delta_graph",
    "SigmaMode",
    "add_edges",
    "add_graph",
    "add_meta",
    "add_nodes",
    "increment_meta",
    "initialize_meta",
    "sigma_graph",
    "EdgeDescriptor",
    "EdgeDescriptorLike",
    "NodeDescriptor",
    "NodeDescriptorLike",
    "TagMode",
    "tag",
]



"""
# Similarity analyses

Graph similarity is often expressed as a metric, where nodes and edges are scanned for similar
patterns, properties, or other aspects. There are three levels of equivalence, being structural,
automorphic, or regular equivalence. Where each of the former implies all latter equivalences,
respectively.

## Available analyses

The following algorithms are directly accessible after importing
[`ragraph.analysis.similarity`][ragraph.analysis.similarity]:

* [`jaccard_index`][ragraph.analysis.similarity.jaccard_matrix]: Jaccard Similarity Index of two
  objects based on the number properties they both possess divided by the number of properties
  either of them have.
* [`jaccard_matrix`][ragraph.analysis.similarity.jaccard_matrix]: Jaccard Similarity Index between a
  set of objects stored in a square matrix.

Note:
    Both Jaccard methods require a callable that takes an object and returns a list of booleans
    representing the possession of a property (the `on` argument). Some examples are included in the
    [`ragraph.analysis.similarity.utils`][ragraph.analysis.similarity.utils] module, like
    [`ragraph.analysis.similarity.utils.on_hasattrs`][ragraph.analysis.similarity.utils.on_hasattrs].
"""

from ragraph.analysis.similarity._jaccard import jaccard_index, jaccard_matrix
from ragraph.analysis.similarity._similarity import SimilarityAnalysis

__all__ = ["jaccard_index", "jaccard_matrix", "SimilarityAnalysis"]



"""# Bus detection algorithms

Bus detection algorithms find bus nodes in graphs, also called "hubs" or "integrative
components". Typical bus nodes have a high node degree or "centrality" score.

All bus detection algorithms have the following arguments:

- `graph`: Graph to find bus nodes in.
- `root`: Root node to perform bus detection in.
- `leafs`: Optional list of leaf nodes to consider during the bus detection cycle. If
  not supplied, the leaf node descendants of the root will be considered.

They always return two lists of nodes:

- Leaf nodes that have been marked as bus nodes.
- The remaining leaf nodes.

They currently do **NOT** change anything in the graph in-place. That is up to the user.

## Available algorithms

The following algorithms are directly available after importing
[`ragraph.analysis.bus`][ragraph.analysis.bus]:

- [`gamma`][ragraph.analysis.bus.gamma]: Gamma bus detection.
"""

from ragraph.analysis.bus._gamma import gamma_bus_detection as gamma

__all__ = ["gamma"]



"""
# Tarjan's strongly connected components algorithm.

Reference:
    Tarjan, R. (1972). Depth-First Search and Linear Graph Algorithms.
    SIAM Journal on Computing, 1(2), 146-160.
    [DOI: 10.1137/0201010] (https://doi.org/10.1137/0201010)
"""
from textwrap import dedent
from typing import Dict, List, Optional, Tuple, Union

from ragraph.analysis._classes import ClusterAnalysis
from ragraph.analysis._utils import create_parent
from ragraph.graph import Graph
from ragraph.node import Node

tarjans_scc_analysis = ClusterAnalysis(
    "Tarjan's strongly connected components clustering algorithm.",
    description=dedent(
        """Clusters all strongly connected components (cycles) in a graph.
        Ignores the `loops` and `edge_weights` parameters."""
    ),
)


@tarjans_scc_analysis
def tarjans_scc(
    graph: Graph,
    root: Optional[Union[str, Node]] = None,
    leafs: Optional[Union[List[Node], List[str]]] = None,
    inherit: bool = True,
    loops: bool = False,
    edge_weights: Optional[List[str]] = None,
    inplace: bool = True,
    names: bool = False,
    safe: bool = True,
    **kwargs,
) -> Tuple[Graph, Union[List[Node], List[str]]]:
    """docstring stub"""
    assert leafs is not None
    clusters = tarjans_scc_algorithm(graph, leafs, inherit)  # type: ignore

    cluster_roots = []
    for children in clusters:
        if len(children) > 1:
            parent = create_parent(graph, children)
            cluster_roots.append(parent)
        else:
            cluster_roots.append(children[0])

    return graph, cluster_roots


def tarjans_scc_algorithm(graph: Graph, nodes: List[Node], inherit: bool) -> List[List[Node]]:
    """Tarjan's strongly connected components algorithm.

    Arguments:
        graph: Graph to detect SCC's in (cycles).
        nodes: List of nodes (components) to detect SCC's for.
        inherit: Whether to take into account (inherit) edges between children during
            calculations.

    Returns:
        Lists of strongly connected components.

    Reference:
        Tarjan, R. (1972). Depth-First Search and Linear Graph Algorithms.
            SIAM Journal on Computing, 1(2), 146-160. https://doi.org/10.1137/0201010
    """
    out = []
    index = 0
    indexes: Dict[Node, int] = {n: -1 for n in nodes}
    lowlinks: Dict[Node, int] = {n: -1 for n in nodes}
    onstacks: Dict[Node, int] = {n: -1 for n in nodes}
    stack = []

    targets_of: Dict[Node, List[Node]] = {
        source: [
            target
            for target in nodes
            if any(graph.edges_between(source, target, inherit=inherit, loops=False))
        ]
        for source in nodes
    }

    def strongconnect(n: Node):
        nonlocal index
        indexes[n] = index
        lowlinks[n] = index
        index = index + 1
        stack.append(n)
        onstacks[n] = True

        for m in targets_of[n]:
            if indexes[m] == -1:
                strongconnect(m)
                lowlinks[n] = min(lowlinks[n], lowlinks[m])  # type: ignore
            elif onstacks[m]:
                lowlinks[n] = min(lowlinks[n], indexes[m])  # type: ignore

        if lowlinks[n] == indexes[n]:
            scc = []
            k = None
            while n != k:
                k = stack.pop()
                onstacks[k] = False
                scc.append(k)
            out.append(scc)

    for node in nodes:
        if indexes[node] == -1:
            strongconnect(node)

    return out



"""XML format support using The Open Group ArchiMate® Model Exchange File Format.

Reference:
    https://www.opengroup.org/xsd/archimate/
"""
import shutil
from pathlib import Path
from typing import Any, Dict, Generator, Optional, Union

from lxml import etree
from lxml.builder import ElementMaker

from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.io.xml import _get_xmi_id
from ragraph.node import Node

here = Path(__file__).parent
schema_doc = etree.parse(str(here / "archimate3_Model.xsd"))
schema = etree.XMLSchema(schema_doc)
schema_root = schema_doc.getroot()

nsmap = {
    None: "http://www.opengroup.org/xsd/archimate/3.0/",
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",
    "xml": "http://www.w3.org/XML/1998/namespace",
}
archi = nsmap[None]
xsi = nsmap["xsi"]
xml = nsmap["xml"]
maker = ElementMaker(nsmap=nsmap)

q_type = etree.QName(xsi, "type")
q_lang = etree.QName(xml, "lang")

annot_types = {bool: "boolean", int: "number", float: "number", str: "string"}


def _make_property_definitions(g: Graph) -> etree.Element:
    """Make a propertyDefinitions element to include in the Archimate XML file.

    Arguments:
        g: Graph to inspect and create required property fields for.

    Returns:
        Property definitions XML element.

    Note:
        Labels, Weight and Annotations all get their own prefixed properties.
    """
    propdefs = maker.propertyDefinitions(
        maker.propertyDefinition(maker.name("kind"), {"identifier": "kind", "type": "string"}),
        maker.propertyDefinition(maker.name("is_bus"), {"identifier": "is_bus", "type": "boolean"}),
    )

    for label in sorted(set(g.node_labels + g.edge_labels)):
        propdefs.append(
            maker.propertyDefinition(
                maker.name(f"label: {label}"),
                dict(identifier=f"_l_{label}", type="boolean"),
            )
        )

    for weight in sorted(set(g.node_weight_labels + g.edge_weight_labels)):
        propdefs.append(
            maker.propertyDefinition(
                maker.name(f"weight: {weight}"),
                dict(identifier=f"_w_{weight}", type="number"),
            )
        )

    annots = set.union(
        {
            (k, annot_types[type(v)])
            for n in g.nodes
            for k, v in n.annotations.items()
            if type(v) in annot_types
        },
        {
            (k, annot_types[type(v)])
            for e in g.edges
            for k, v in e.annotations.items()
            if type(v) in annot_types
        },
    )

    for annot in annots:
        propdefs.append(
            maker.propertyDefinition(
                maker.name(f"annotation({annot[1]}): {annot[0]}"),
                dict(identifier=f"_a_{annot[1]}_{annot[0]}", type=annot[1]),
            )
        )

    return propdefs


def _get_properties(obj: Union[Node, Edge]) -> etree.Element:
    """Get the properties element for an object being a Node or Edge.

    Arguments:
        obj: Node or Edge to fetch the properties element for.

    Returns:
        Properties XML element.
    """
    props = maker.properties()

    props.append(maker.property(maker.value(obj.kind), propertyDefinitionRef="kind"))

    for i in sorted(obj.labels):
        props.append(maker.property(maker.value("true"), propertyDefinitionRef=f"_l_{i}"))

    for k in sorted(obj.weights.keys()):
        v = obj.weights[k]
        props.append(maker.property(maker.value(str(float(v))), propertyDefinitionRef=f"_w_{k}"))

    for k in sorted(obj.annotations.keys()):
        v = getattr(obj.annotations, k)
        if type(v) not in annot_types:
            continue
        t = annot_types[type(v)]
        props.append(maker.property(maker.value(str(v)), propertyDefinitionRef=f"_a_{t}_{k}"))

    if isinstance(obj, Node) and obj.is_bus:
        props.append(maker.property(maker.value("true"), propertyDefinitionRef="is_bus"))

    return props


def _node_to_element(node: Node) -> etree.Element:
    """Convert a Node into an XML element.

    Arguments:
        node: Node to convert into an XML element.

    Returns:
        Node XML element.

    Note:
        node.annotations.archimate["type"] determines the type (and therefore layer).
        node.annotations.archimate["documentation"] populates the documentation field.
    """
    # Archimate element name.
    archi_name = maker.name(node.name, {q_lang: "en"})

    # Rudimentary element fields including Archimate element type.
    # Type is determined from the "archimate" annotation:
    archi_type = getattr(node.annotations, "archimate", {}).get("type", "Node")
    archi_fields = {"identifier": _get_xmi_id(node, "_n_"), q_type: archi_type}

    # Archimate element properties (kind/labels/weights/annotations with a value)
    archi_properties = _get_properties(node)

    # Get documentation field if the respective annotation is found.
    documentation = getattr(node.annotations, "archimate", {}).get("documentation", None)

    # Make element with or without documentation.
    if documentation is None:
        elem = maker.element(archi_name, archi_properties, archi_fields)
    else:
        archi_docs = maker.documentation(str(documentation).replace("\n", "\r\n"), {q_lang: "en"})
        elem = maker.element(archi_name, archi_docs, archi_properties, archi_fields)

    return elem


def _hierarchy_to_relationships(graph: Graph) -> Generator[etree.Element, None, None]:
    """Generate relationships of a Graph's Node hierarchy.

    Arguments:
        graph: Graph to generate hierarchical relationship of.

    Yields:
        Composition relationship XML elements.
    """
    for node in graph.nodes:
        if node.parent is None:
            continue
        parent_id = _get_xmi_id(node.parent, "_n_")
        child_id = _get_xmi_id(node, "_n_")

        yield maker.relationship(
            maker.name(f"{node.parent.name} contains {node.name}", {q_lang: "en"}),
            {
                "source": parent_id,
                "target": child_id,
                "identifier": f"_h_{parent_id}_{child_id}",
                q_type: "Composition",
            },
        )


def _edge_to_relationship(edge: Edge) -> etree.Element:
    """Convert an Edge into an XML element.

    Arguments:
        edge: Edge to convert into an XML relationship element.

    Returns:
        Relationship XML element.

    Note:
        edge.annotations.archimate["type"] determines the type (and therefore layer).
        edge.annotations.archimate["documentation"] populates the documentation field.
    """
    # Archimate element name.
    archi_name = maker.name(f"{edge.source.name} to {edge.target.name}", {q_lang: "en"})

    # Rudimentary element fields including Archimate element type.
    # Type is determined from the "archimate" annotation:
    archi_type = getattr(edge.annotations, "archimate", {}).get("type", "Association")
    archi_fields = {
        "source": _get_xmi_id(edge.source, "_n_"),
        "target": _get_xmi_id(edge.target, "_n_"),
        "identifier": _get_xmi_id(edge, "_e_"),
        q_type: archi_type,
    }

    # Archimate element properties (kind/labels/weights/annotations with a value)
    archi_properties = _get_properties(edge)

    # Get documentation field if the respective annotation is found.
    documentation = getattr(edge.annotations, "archimate", {}).get("documentation", None)

    # Make element with or without documentation.
    if documentation is None:
        elem = maker.relationship(
            archi_name,
            archi_properties,
            archi_fields,
        )
    else:
        archi_docs = maker.documentation(str(documentation).replace("\n", "\r\n"), {q_lang: "en"})
        elem = maker.relationship(archi_name, archi_docs, archi_properties, archi_fields)

    return elem


def _expand_requirements_constraints(node: Node, elems: etree.Element, rels: etree.Element):
    """Expand requirements and constraints so that they become an ApplicationFunction
    and Requirement/Constraint combination.

    Arguments:
        node: Nodes to expand.
        elems: XML element to append elements to.
        rels: XML element to append relationships to.

    Note:
        Looks for annotations.archimate["requirement"] or
            annotations.archimate["constraint"] and builds additional elements when
            found.
    """
    # Check whether constraint/requirement field exists and get content.
    annot = getattr(node.annotations, "archimate", {})
    if "constraint" in annot:
        archi_type = "Constraint"
        contents = annot["constraint"]
    elif "requirement" in annot:
        archi_type = "Requirement"
        contents = annot["requirement"]
    else:
        return

    # Append Requirement/Constraint element.
    xmi_id = _get_xmi_id(node, "_r_")
    archi_name = maker.name(node.name, {q_lang: "en"})
    archi_fields = {"identifier": xmi_id, q_type: archi_type}
    archi_docs = maker.documentation(contents, {q_lang: "en"})
    elems.append(maker.element(archi_name, archi_fields, archi_docs))

    # Append Realization edge.
    archi_name = maker.name(f"Realization of {node.name}", {q_lang: "en"})
    archi_fields = {
        "source": _get_xmi_id(node, "_n_"),
        "target": xmi_id,
        "identifier": _get_xmi_id(node, "_e_"),
        q_type: "Realization",
    }
    rels.append(maker.relationship(archi_name, archi_fields))


def to_archimate(
    graph: Graph,
    path: Optional[Union[str, Path]] = None,
    elem: Optional[etree.Element] = None,
    tostring_args: Optional[Dict[str, Any]] = None,
    bundle_schemas: bool = False,
) -> str:
    """Encode Graph to an Archimate model XML element.

    Arguments:
        graph: Graph to convert to XML.
        path: Optional file path to write XML to.
        elem: Optional object to append the Graph to.
        tostring_args: Optional argument overrides to
            [`lxml.etree.tostring`][lxml.etree.tostring].
        bundle_schemas: When exporting to a file, bundle the .xsd schemas.

    Returns:
        XML element.

    Note:
        For both nodes and edges we use the following:
        The `annotations.archimate["type"]` determines the type (and therefore layer).
        The `annotations.archimate["documentation"]` populates the documentation field.
        Refer to https://www.opengroup.org/xsd/archimate/ and the XSD scheme for all
        possible XML element types.
    """
    if elem is None:
        parser = etree.XMLParser(remove_blank_text=True)
        doc = etree.parse(str(here / "bare.xml"), parser)
        root = doc.getroot()
    else:
        root = elem

    elements = maker.elements()
    elements.extend(_node_to_element(n) for n in graph.nodes)
    if len(elements):
        root.append(elements)

    relationships = maker.relationships()
    relationships.extend(_hierarchy_to_relationships(graph))
    relationships.extend(_edge_to_relationship(e) for e in graph.edges)
    if len(relationships):
        root.append(relationships)

    propdefs = _make_property_definitions(graph)
    if len(propdefs):
        root.append(propdefs)

    for n in graph.nodes:
        _expand_requirements_constraints(n, elements, relationships)

    if path is not None:
        p = Path(path)

        args = dict(encoding="UTF-8", xml_declaration=True, pretty_print=True)
        if tostring_args is not None:
            args.update(tostring_args)

        string = etree.tostring(root, **args)
        p.write_bytes(string)

        if bundle_schemas:
            shutil.copy2(here / "archimate3_Model.xsd", p.parent)

    return root


def from_archimate(*args, **kwargs):
    """Archimate XML import is not implemented."""
    raise NotImplementedError("Archimate XML import is not implemented.")



"""
# Clustering algorithms

Clustering algorithms detect (nested) clusters of nodes in graphs that are relatively
tightly connected by means of their edges. Both hierarchical and flat clustering
algorithms are provided.

Apart of algorithm specific parameters, all of them feature the same basic parameters:

- `graph`: Graph to cluster containing the relevant nodes and edges.
- `leafs`: Optional list of leaf nodes to cluster. If not provided all the graph's
  leaf nodes are selected.
- `inplace`: Boolean toggle whether to create the new cluster nodes in the provided
  graph or provided a deepcopy of the graph with only the leaf nodes, their edges and
  newly created cluster nodes.

## Available algorithms

- [`markov`][ragraph.analysis.cluster.markov]
- [`hierarchical_markov`][ragraph.analysis.cluster.hierarchical_markov]
- ['tarjans_scc`][ragraph.analysis.cluster.tarjans_scc]
"""

from ragraph.analysis.cluster._markov import (
    MarkovFlow,
    MarkovRelative,
    calculate_tpm,
    create_clusters,
    get_sink_matrix,
    hierarchical_markov,
    markov,
    prune_matrix,
)
from ragraph.analysis.cluster._tarjan import tarjans_scc

__all__ = [
    "hierarchical_markov",
    "markov",
    "tarjans_scc",
    "calculate_tpm",
    "create_clusters",
    "create_parent",
    "get_sink_matrix",
    "MarkovFlow",
    "MarkovRelative",
    "prune_matrix",
]



"""# Blank plot component"""

import plotly.graph_objs as go

from ragraph.plot.generic import Component, Style


class Blank(Component):
    """Blank plot component.

    Arguments:
        style: Plot style mapping.
    """

    def __init__(self, style: Style = Style()):
        trace = go.Scatter(
            x=[0.5 * style.xstep],
            y=[0.5 * style.ystep],
            mode="markers",
            marker=dict(
                line=dict(
                    color="#FFFFFF",
                ),
                color="#FFFFFF",
            ),
            showlegend=False,
            hoverinfo="skip",
        )

        xaxis = go.layout.XAxis(
            automargin=False,
            autorange=False,
            scaleanchor="y",
            scaleratio=1.0,
            range=(0, 1),
            showticklabels=False,
        )
        yaxis = go.layout.YAxis(
            automargin=False,
            autorange=False,
            showticklabels=False,
            range=(0, 1),
        )

        super().__init__(
            width=style.xstep * style.boxsize,
            height=style.ystep * style.boxsize,
            traces=[trace],
            xaxis=xaxis,
            yaxis=yaxis,
        )



"""# Graph handling utilities"""
from typing import TYPE_CHECKING, Generator, List, Set

if TYPE_CHECKING:
    from ragraph.graph import Graph
    from ragraph.node import Node


def select_nodes(
    graph: "Graph", node_kinds: List[str], edge_kinds: List[str], depth: int, selection_mode: str
) -> List["Node"]:
    """Select specific nodes from this graph in a structured order.

    Arguments:
        graph: Graph to fetch data from.
        node_kinds: The kind of nodes to be selected.
        edge_kinds: The kind of edges to be selected.
        depth: The maximum depth of node to be selected.
        selection_mode: The selection mode. Either 'dependent' or 'independent'.

    Note:
        The selection mode argument determines how nodes of different kinds are
        selected. If the selection mode is set to 'dependent', the first node kind
        in the `node_kinds` list is considered to be the 'lead node kind'.
        Nodes of different kind than the lead node kind, are only selected if they
        have a dependency with at least one of the selected nodes of the lead node
        kind. If the selection mode is set to 'independent' this dependency
        condition is dropped.
    """
    node_kinds = node_kinds or graph.node_kinds

    if selection_mode not in ["dependent", "independent"]:
        raise ValueError(
            f'Unrecognized selection mode "{selection_mode}", '
            'should be "independent" or "dependent".'
        )

    node_kind_order = {kind: i for i, kind in enumerate(node_kinds)}
    roots = sorted(
        [n for n in graph.roots if n.kind in node_kinds],
        key=lambda n: node_kind_order[n.kind],
    )

    nodes = list(get_up_to_depth(roots, depth))

    if selection_mode == "independent" or len(node_kinds) == 1:
        return nodes

    lead_kind = node_kinds[0]
    i = 0
    while i < len(nodes) and nodes[i].kind == lead_kind:
        i += 1
    lead_nodes, candidate_nodes = nodes[:i], nodes[i:]

    edge_kinds = edge_kinds or graph.edge_kinds
    edge_kindset = set(edge_kinds)
    dependent_nodes = [
        n for n in candidate_nodes if is_dependent(graph, lead_nodes, n, edge_kindset)
    ]

    return lead_nodes + dependent_nodes


def get_up_to_depth(roots: List["Node"], depth: int) -> Generator["Node", None, None]:
    """Get nodes up to a certain depth with bus nodes at the start of child lists."""
    for node in roots:
        if node.is_leaf or node.depth == depth:
            yield node
            continue

        children = sorted(node.children, key=lambda n: n.is_bus, reverse=True)

        if node.depth == depth - 1:
            yield from children
        else:
            yield from get_up_to_depth(children, depth)


def is_dependent(graph, lead_nodes: List["Node"], node: "Node", edge_kinds=Set[str]) -> bool:
    """Check if a node is dependent on the lead node kind.

    E.g. an edge of the allowed kinds exists to or from the lead nodes.
    """
    return any(
        True for e in graph.edges_between_all(lead_nodes, [node]) if e.kind in edge_kinds
    ) or any(True for e in graph.edges_between_all([node], lead_nodes) if e.kind in edge_kinds)



"""# Comparison analysis utilities"""
import enum
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import FrozenSet, Tuple

from ragraph.edge import Edge
from ragraph.generic import Metadata
from ragraph.graph import Graph
from ragraph.node import Node


class NodeDescriptorLike(ABC):
    """A class that takes a node and describes it by returning a fixed (hashable) output.

    Hashable means that the description can be used to determine uniqueness by set operations.
    """

    @classmethod
    @abstractmethod
    def from_node(cls, node: Node) -> "NodeDescriptorLike":
        """Create a node descriptor from a node."""
        raise NotImplementedError

    @abstractmethod
    def to_node(self, graph: Graph) -> Node:
        """Create a node from a descriptor in a graph and return it for convenience."""
        raise NotImplementedError


@dataclass(frozen=True)
class NodeDescriptor(NodeDescriptorLike):
    """Describe a node based on its name, kind, labels and weights."""

    name: str
    kind: str
    labels: FrozenSet[str]
    weights: FrozenSet[Tuple[str, float]]

    @classmethod
    def from_node(cls, node: Node) -> "NodeDescriptor":
        return cls(
            node.name,
            node.kind,
            frozenset(node.labels),
            frozenset(node.weights.items()),
        )

    def to_node(self, graph: Graph) -> Node:
        node = Node(
            name=self.name,
            kind=self.kind,
            labels=list(self.labels),
            weights=dict(self.weights),
        )
        graph.add_node(node)
        return node


class EdgeDescriptorLike(ABC):
    """A class that takes an edge and describes it by returning a fixed (hashable) output.

    Hashable means that the description can be used to determine uniqueness by set operations.
    """

    @classmethod
    @abstractmethod
    def from_edge(cls, edge: Edge) -> "EdgeDescriptor":
        """Create an edge descriptor from an edge. Ignores annotations."""
        ...

    @abstractmethod
    def to_edge(self, graph: Graph) -> Edge:
        """Create an edge from a descriptor in a graph."""
        ...


@dataclass(frozen=True)
class EdgeDescriptor(EdgeDescriptorLike):
    """Describe a node based on its source name, target name, kind, labels, and weights."""

    source: str
    target: str
    kind: str
    labels: FrozenSet[str]
    weights: FrozenSet[Tuple[str, float]]

    @classmethod
    def from_edge(cls, edge: Edge) -> "EdgeDescriptor":
        return EdgeDescriptor(
            edge.source.name,
            edge.target.name,
            edge.kind,
            frozenset(edge.labels),
            frozenset(edge.weights.items()),
        )

    def to_edge(self, graph: Graph) -> Edge:
        edge = Edge(
            graph[self.source],
            graph[self.target],
            kind=self.kind,
            labels=list(self.labels),
            weights=dict(self.weights),
        )
        graph.add_edge(edge)
        return edge


class TagMode(enum.StrEnum):
    """How to tag nodes as unique to one of the given inputs or common."""

    KIND = enum.auto()
    LABEL = enum.auto()
    ANNOTATION = enum.auto()


def tag(item: Metadata, tag: str, mode: TagMode):
    """Tag an item using a given tagging mode."""
    if mode == TagMode.KIND:
        item.kind = tag
    elif mode == TagMode.LABEL:
        item.labels = item.labels + [tag]
    elif mode == TagMode.ANNOTATION:
        item.annotations[tag] = True
    else:
        raise ValueError("Unknown tagging mode.")



"""# RaGraph generic plotting classes"""
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from plotly import graph_objs as go
from plotly.basedatatypes import BaseTraceType

from ragraph import colors
from ragraph.generic import Mapping, field
from ragraph.plot import svg

MODEBAR_BUTTONS = {
    "toImage",
    "sendDataToCloud",
    "editInChartStudio",
    "zoom2d",
    "pan2d",
    "select2d",
    "lasso2d",
    "zoomIn2d",
    "zoomOut2d",
    "autoScale2d",
    "resetScale2d",
    "hoverClosestCartesian",
    "hoverCompareCartesian",
    "zoom3d",
    "pan3d",
    "orbitRotation",
    "tableRotation",
    "resetCameraDefault3d",
    "resetCameraLastSave3d",
    "hoverClosest3d",
    "zoomInGeo",
    "zoomOutGeo",
    "resetGeo",
    "hoverClosestGeo",
    "hoverClosestGl2d",
    "hoverClosestPie",
    "resetSankeyGroup",
    "toggleHover",
    "resetViews",
    "toggleSpikelines",
    "resetViewMapbox",
}


class TreeStyle(Mapping):
    """RaGraph tree plot component style mapping.

    Arguments:
        line: Line style mapping for hierarchy tree lines.
        xaxis: Plotly X-axis settings.
        yaxis: Plotly Y-axis settings.
    """

    _defaults = dict(
        line=svg.Line(color="gray"),
        xaxis=go.layout.XAxis(
            showgrid=False,
            showticklabels=False,
            automargin=False,
            scaleratio=1.0,
            zeroline=False,
            showline=False,
            autorange=False,
        ),
        yaxis=go.layout.YAxis(
            autorange=False,
            scaleratio=1.0,
            automargin=False,
            showgrid=True,
            showticklabels=False,
            scaleanchor="x",
            zeroline=False,
            showline=False,
        ),
    )

    def __init__(
        self,
        line: Optional[Union[svg.Line, Dict[str, Any]]] = None,
        xaxis: Optional[Union[go.layout.XAxis, Dict[str, Any]]] = None,
        yaxis: Optional[Union[go.layout.YAxis, Dict[str, Any]]] = None,
    ):
        if isinstance(xaxis, dict):
            _xaxis = deepcopy(self._defaults["xaxis"])
            for k, v in xaxis.items():
                setattr(_xaxis, k, v)
            xaxis = _xaxis

        if isinstance(yaxis, dict):
            _yaxis = deepcopy(self._defaults["yaxis"])
            for k, v in yaxis.items():
                setattr(_yaxis, k, v)
            yaxis = _yaxis

        super().__init__(line=line, xaxis=xaxis, yaxis=yaxis)

    @field
    def line(self) -> svg.Line:  # type: ignore
        """Line style mapping for hierarchy tree lines."""

    @field
    def xaxis(self) -> go.layout.XAxis:  # type: ignore
        """Plotly X-axis settings."""

    @field
    def yaxis(self) -> go.layout.YAxis:  # type: ignore
        """Plotly Y-axis settings."""


class LabelsStyle(Mapping):
    """Labels plot component style mapping.

    Arguments:
        fontcolor: Font color used for labels.
        fontfamily: Font family used for labels.
        fontsize: Fontsize used for labels.
        fontaspectratio: Font width per fontsize ratio.
        textorientation: Orientation of label text, one of "horizontal", "vertical".
        xaxis: Plotly X-axis settings.
        yaxis: Plotly Y-axis settings.
    """

    _defaults = dict(
        fontcolor="black",
        fontfamily="Hack,Courier New,monospace",
        fontsize=None,
        fontaspectratio=0.6005,
        shorten=True,
        textorientation="horizontal",
        xaxis=go.layout.XAxis(
            automargin=False,
            scaleanchor="y",
            showgrid=False,
            showline=False,
            showticklabels=False,
            zeroline=False,
        ),
        yaxis=go.layout.YAxis(
            automargin=False,
            autorange=False,
            scaleratio=1.0,
            showgrid=False,
            showline=False,
            showticklabels=False,
            zeroline=False,
        ),
    )

    def __init__(
        self,
        fontcolor: Optional[str] = None,
        fontfamily: Optional[str] = None,
        fontsize: Optional[int] = None,
        fontaspectratio: Optional[float] = None,
        shorten=None,
        textorientation: Optional[str] = None,
        xaxis: Optional[Union[go.layout.XAxis, Dict[str, Any]]] = None,
        yaxis: Optional[Union[go.layout.YAxis, Dict[str, Any]]] = None,
    ):
        if isinstance(xaxis, dict):
            _xaxis = deepcopy(self._defaults["xaxis"])
            for k, v in xaxis.items():
                setattr(_xaxis, k, v)
            xaxis = _xaxis

        if isinstance(yaxis, dict):
            _yaxis = deepcopy(self._defaults["yaxis"])
            for k, v in yaxis.items():
                setattr(_yaxis, k, v)
            yaxis = _yaxis

        super().__init__(
            fontcolor=fontcolor,
            fontfamily=fontfamily,
            fontsize=fontsize,
            fontaspectratio=fontaspectratio,
            shorten=shorten,
            textorientation=textorientation,
            xaxis=xaxis,
            yaxis=yaxis,
        )

    @field
    def fontcolor(self) -> str:  # type: ignore
        """Font color used for labels."""

    @field
    def fontfamily(self) -> str:  # type: ignore
        """Font family used for labels."""

    @field
    def fontsize(self) -> int:  # type: ignore
        """Fontsize used for labels."""

    @field
    def fontaspectratio(self) -> float:  # type: ignore
        """Font width per fontsize ratio."""

    @field
    def shorten(self) -> Union[bool, Callable[[str], str]]:  # type: ignore
        """Label shortening toggle or function. When set to `True` everything after
        the last dot '.' is kept.
        """

    @field
    def textorientation(self) -> float:  # type: ignore
        """Orientation of text."""

    @field
    def xaxis(self) -> go.layout.XAxis:  # type: ignore
        """Plotly X-axis settings."""

    @field
    def yaxis(self) -> go.layout.YAxis:  # type: ignore
        """Plotly Y-axis settings."""


class PieMapStyle(Mapping):
    """Piechart map's plot component style mapping.

    Arguments:
        busarea: Bus area SVG mapping. Used for styling the bus area.
        display: What to display. One of 'kinds', 'labels', 'weight labels', 'weights'.
        fields: The fields to plot (the selection of kinds, labels, or weights). Leave set to `None`
            to display all the available fields automatically.
        gridline: Grid line options mapping.
        highlight_col_annotation: Annotation that signals what columns should be highlighted.
            Value should be True-ish.
        highlight_col_color: Default color to use for column highlights.
        highlight_row_annotation: Annotation that signals what rows should be highlighted.
            Value should be True-ish.
        highlight_row_color: Default color to use for row highlights.
        inherit: Whether to display edges between descendants of the axis nodes.
        kindline: Node kind separation lines options mapping.
        mode: How to divide the pie-charts per category: 'equal' or 'relative'.
            'equal' divides the piechart evenly.
            'relative' divides the piechart according to category value.
        radius: The piechart radius between 0.0 and 0.5.
        scale_weight: Edge weight label that should contain values between 0.0 and 1.0 to scale the
            radius with.
        customhoverkeys: List of keys for information to be displayed on hover.
        xaxis: Plotly X-axis settings.
        yaxis: Plotly Y-axis settings.

    Note:
        The `display` argument determines what is going to be plotted as piecharts in the plot area.
        The `fields` argument is a filter on the possible values for that display mode. The `mode`
        argument then tunes how the wedges that make up the piecharts should be distributed. Most of
        the time, 'equal' gives the most predictable and clear results.
    """

    _defaults = dict(
        busarea=svg.SVG(fillcolor="rgba(150,150,150,0.2)", line=svg.Line(width=0), layer="below"),
        clusterline=svg.Line(width=2, color="gray", dash="solid"),
        display="kinds",
        fields=None,
        gridline=svg.Line(color="#dfdfdf", width=2, dash="solid"),
        highlight_col_annotation=None,
        highlight_col_color=None,
        highlight_row_annotation=None,
        highlight_row_color=None,
        inherit=False,
        kindline=svg.Line(color="gray", width=2, dash="dot"),
        mode="equal",
        radius=0.4,
        scale_weight=None,
        customhoverkeys=[],
        xaxis=go.layout.XAxis(
            automargin=False,
            scaleanchor="y",
            showgrid=False,
            showline=False,
            showticklabels=False,
            zeroline=False,
        ),
        yaxis=go.layout.YAxis(
            automargin=False,
            autorange=False,
            scaleratio=1.0,
            showgrid=False,
            showline=False,
            showticklabels=False,
            zeroline=False,
        ),
    )

    def __init__(
        self,
        busarea: Optional[svg.SVG] = None,
        clusterline: Optional[svg.Line] = None,
        display: Optional[str] = None,
        fields: Optional[List[str]] = None,
        gridline: Optional[svg.Line] = None,
        highlight_col_annotation: Optional[str] = None,
        highlight_col_color: Optional[str] = None,
        highlight_row_annotation: Optional[str] = None,
        highlight_row_color: Optional[str] = None,
        inherit: Optional[bool] = None,
        kindline: Optional[svg.Line] = None,
        mode: Optional[str] = None,
        radius: Optional[float] = None,
        scale_weight: Optional[str] = None,
        customhoverkeys: Optional[List[str]] = None,
        xaxis: Optional[Dict[str, Any]] = None,
        yaxis: Optional[Dict[str, Any]] = None,
    ):
        if isinstance(xaxis, dict):
            _xaxis = deepcopy(self._defaults["xaxis"])
            for k, v in xaxis.items():
                setattr(_xaxis, k, v)
            xaxis = _xaxis

        if isinstance(yaxis, dict):
            _yaxis = deepcopy(self._defaults["yaxis"])
            for k, v in yaxis.items():
                setattr(_yaxis, k, v)
            yaxis = _yaxis

        super().__init__(
            busarea=busarea,
            clusterline=clusterline,
            display=display,
            fields=fields,
            gridline=gridline,
            highlight_col_annotation=highlight_col_annotation,
            highlight_col_color=highlight_col_color,
            highlight_row_annotation=highlight_row_annotation,
            highlight_row_color=highlight_row_color,
            inherit=inherit,
            kindline=kindline,
            mode=mode,
            radius=radius,
            scale_weight=scale_weight,
            customhoverkeys=customhoverkeys,
            xaxis=xaxis,
            yaxis=yaxis,
        )

    @field
    def busarea(self) -> svg.SVG:  # type: ignore
        """Bus area SVG mapping. Used for styling the bus area."""

    @field
    def display(self) -> str:  # type: ignore
        """What to display. One of 'kinds', 'labels', 'weight labels', 'weights'."""

    @field
    def fields(self) -> List[str]:  # type: ignore
        """The fields to plot (the selection of kinds, labels, or weights). Leave set
        to `None` to display all the available fields automatically.
        """

    @field
    def gridline(self) -> svg.Line:  # type: ignore
        """Grid line style."""

    @field
    def highlight_col_annotation(self) -> Optional[str]:  # type: ignore
        """Annotation that signals what columns should be highlighted.
        Value should be True-ish.
        """

    @field
    def highlight_col_color(self) -> Optional[str]:  # type: ignore
        """Default color to use for column highlights."""

    @field
    def highlight_row_annotation(self) -> Optional[str]:  # type: ignore
        """Annotation that signals what rows should be highlighted.
        Value should be True-ish.
        """

    @field
    def highlight_row_color(self) -> Optional[str]:  # type: ignore
        """Default color to use for row highlights."""

    @field
    def inherit(self) -> bool:  # type: ignore
        """Whether to display edges between descendants of the axis nodes."""

    @field
    def kindline(self) -> svg.Line:  # type: ignore
        """Node kind separation lines options mapping."""

    @field
    def mode(self) -> str:  # type: ignore
        """How to divide the piecharts per field. Either 'equal' or 'relative'.
        'equal' divides the piecharts evenly.
        'relative' divides the piechart according to field value.
        """

    @field
    def radius(self) -> float:  # type: ignore
        """The piechart radius between 0.0 and 0.5."""

    @field
    def scale_weight(self) -> str:  # type: ignore
        """Edge weight label that should contain values between 0.0 and 1.0 to scale the
        radius with.
        """

    @field
    def customhoverkeys(self) -> List[str]:  # type: ignore
        """Custom keys for information to be displayed on hover."""

    @field
    def xaxis(self) -> go.layout.XAxis:  # type: ignore
        """Plotly X-axis settings."""

    @field
    def yaxis(self) -> go.layout.YAxis:  # type: ignore
        """Plotly Y-axis settings."""


class FieldPalette(Mapping):
    """Palettes for a field in a plot.

    Argument:
        categorical: Categorical color for this field.
        continuous: Continuous (numeric) data color palette for this field.
    """

    _defaults = dict(
        categorical=None,
        continuous=None,
    )

    def __init__(
        self,
        categorical: Optional[Union[str, List[str]]] = None,
        continuous: Optional[List[str]] = None,
    ):
        super().__init__(self, categorical=categorical, continuous=continuous)

    @field
    def categorical(self) -> Optional[str]:  # type: ignore
        """Categorical color for this field."""

    @field
    def continuous(self) -> Optional[List[str]]:  # type: ignore
        """Continuous (numeric) data color palette for this field."""


class Palettes(Mapping):
    """Plot palettes mapping.

    Arguments:
        categorical: Categorical data color palette.
        continuous: Continuous (numeric) data color palette.
        fields: Palette override dictionary per display field.
        domains: Value domains to interpolate palettes between per field as a tuple of
            (lower, upper) bounds. Only used for continuous fields.
    """

    _defaults = dict(
        categorical=colors.get_categorical(),
        continuous=colors.get_continuous(),
        fields=dict(),
        domains=dict(),
    )

    def __init__(
        self,
        categorical: Optional[List[str]] = None,
        continuous: Optional[List[str]] = None,
        fields: Optional[Dict[str, FieldPalette]] = None,
        domains: Optional[Dict[str, Tuple[float, float]]] = None,
    ):
        super().__init__(
            categorical=categorical,
            continuous=continuous,
            fields=fields,
            domains=domains,
        )

    def get_categorical_palette(self, field: Optional[str] = None) -> Union[str, List[str]]:
        """Get a categorical color (palette). Might be an overridden color, colorlist,
        or the default palette for the given field."""
        palette = self.fields.get(field, self.categorical)

        # Handle when it's a dictionary/FieldPalette override.
        if isinstance(palette, dict) or isinstance(palette, FieldPalette):
            palette = palette.get("categorical", self.categorical)

        # Failsafe backup.
        if palette is None:
            palette = self.categorical

        return palette

    def get_categorical_color(self, idx: int, field: Optional[str] = None) -> str:
        """Get a color from a categorical palette."""
        palette = self.get_categorical_palette(field=field)

        if isinstance(palette, str):
            return palette
        else:
            return palette[idx]

    def get_continuous_palette(self, field: Optional[str] = None) -> List[str]:
        """Get a continuous color palette."""
        palette = self.fields.get(field, self.continuous)

        # Handle when it's a dictionary/FieldPalette override.
        if isinstance(palette, dict) or isinstance(palette, FieldPalette):
            palette = palette.get("continuous", self.continuous)

        # Failsafe backup.
        return self.continuous if palette is None else palette

    def get_continuous_color(
        self,
        value: float,
        lower: float,
        upper: float,
        field: Optional[str] = None,
    ) -> str:
        """Get a color from the continuous palette by its interpolated index.

        Arguments:
            value: Value to get an interpolated color for.
            lower: Lower bound value (overridden if selected field is in domains).
            upper: Upper bound value (overridden if selected field is in domains).
            field: Optional field to fetch the palette for.
        """
        palette = self.get_continuous_palette(field=field)

        # Override lower and upper if domains are set.
        if field in self.domains:
            lower, upper = self.domains[field]

        # No scale, return highest.
        if lower == upper:
            return palette[-1]

        step = (upper - lower) / (len(palette) - 1)
        idx = int((value - lower) // step)

        idx = min(max(idx, 0), len(palette) - 1)
        return palette[idx]

    @field
    def categorical(self) -> List[str]:  # type: ignore
        """Categorical data color palette."""

    @field
    def continuous(self) -> List[str]:  # type: ignore
        """Continuous (numeric) data color palette."""

    @field
    def fields(self) -> Dict[str, Union[str, List[str], FieldPalette]]:  # type: ignore
        """Palette override dictionary per display field."""

    @field
    def domains(self) -> Dict[str, Tuple[float, float]]:  # type: ignore
        """Value domains to interpolate palettes between per field as a tuple of
        (lower, upper), bounds. Only used for continuous fields.
        """


class LegendStyle(Mapping):
    """Legend plot component style mapping.

    Arguments:
        fontcolor: Font color used for labels.
        fontfamily: Font family used for labels.
        fontsize: Fontsize used for labels.
        fontaspectratio: Font width per fontsize ratio.
        height: Height of the swatch plot in number of box sizes when
            plotting a numerical legend.
        n_ticks: Number of ticks in the swatch plot when plotting a numerical legend.
        xaxis: Plotly X-axis settings.
        yaxis: Plotly Y-axis settings.
    """

    _defaults = dict(
        fontcolor="black",
        fontfamily="Hack,Courier New,monospace",
        fontsize=None,
        fontaspectratio=0.6005,
        height=10,
        n_ticks=5,
        xaxis=go.layout.XAxis(
            automargin=False,
            scaleanchor="y",
            showgrid=False,
            showline=False,
            showticklabels=False,
            zeroline=False,
        ),
        yaxis=go.layout.YAxis(
            automargin=False,
            autorange=False,
            scaleratio=1.0,
            showgrid=False,
            showline=False,
            showticklabels=False,
            zeroline=False,
        ),
    )

    def __init__(
        self,
        fontcolor: Optional[str] = None,
        fontfamily: Optional[str] = None,
        fontsize: Optional[int] = None,
        fontaspectratio: Optional[float] = None,
        height: Optional[int] = None,
        n_ticks: Optional[int] = None,
        xaxis: Optional[Union[go.layout.XAxis, Dict[str, Any]]] = None,
        yaxis: Optional[Union[go.layout.YAxis, Dict[str, Any]]] = None,
    ):
        if isinstance(xaxis, dict):
            _xaxis = deepcopy(self._defaults["xaxis"])
            for k, v in xaxis.items():
                setattr(_xaxis, k, v)
            xaxis = _xaxis

        if isinstance(yaxis, dict):
            _yaxis = deepcopy(self._defaults["yaxis"])
            for k, v in yaxis.items():
                setattr(_yaxis, k, v)
            yaxis = _yaxis

        super().__init__(
            fontcolor=fontcolor,
            fontfamily=fontfamily,
            fontsize=fontsize,
            fontaspectratio=fontaspectratio,
            height=height,
            n_ticks=n_ticks,
            xaxis=xaxis,
            yaxis=yaxis,
        )

    @field
    def fontcolor(self) -> str:  # type: ignore
        """Font color used for labels."""

    @field
    def fontfamily(self) -> str:  # type: ignore
        """Font family used for labels."""

    @field
    def fontsize(self) -> int:  # type: ignore
        """Fontsize used for labels."""

    @field
    def fontaspectratio(self) -> float:  # type: ignore
        """Font width per fontsize ratio."""

    @field
    def height(self) -> int:  # type: ignore
        """Height of the swatch plot in number of box sizes when plotting a
        numerical legend."""

    @field
    def n_ticks(self) -> int:  # type: ignore
        """Number of ticks in the swatch plot when plotting a numerical legend."""

    @field
    def xaxis(self) -> go.layout.XAxis:  # type: ignore
        """Plotly X-axis settings."""

    @field
    def yaxis(self) -> go.layout.YAxis:  # type: ignore
        """Plotly Y-axis settings."""


class Style(Mapping):
    """RaGraph plot style mapping.

    Arguments:
        boxsize: Size in pixels per row or column.
        config: Plotly Figure.show() config.
        highlight_annotation: Annotation key of instances that should be highlighted.
            Value should be True-ish. Set key to `None` to disable.
        highlight_color: Default color to use for highlights.
        labels: Labels plot style.
        layout: Layout options.
        palettes: Plot palettes options.
        piemap: Piechart map plot style.
        tree: Tree plot style.
        legend: Legend plot style.
        show_legend: Bool to display legend.
        row_col_numbers: Bool to display row and column numbers.
        xstep: Axis increment per row or column in plots (usually 1).
        ystep: Axis increment per row or column in plots (usually 1).
    """

    _defaults = dict(
        boxsize=20,
        config=dict(
            displaylogo=False,
            modeBarButtonsToRemove=list(
                MODEBAR_BUTTONS - {"resetScale2d", "toImage", "toggleSpikelines"}
            ),
            responsive=False,
            scrollZoom=False,
        ),
        highlight_annotation="highlight",
        highlight_color="rgba(130, 163, 254, 0.25)",
        labels=LabelsStyle(),
        layout=go.Layout(
            autosize=False,
            margin=go.layout.Margin(l=0, r=0, b=0, t=25, pad=0),
            modebar=dict(orientation="h"),
            hovermode="closest",
            paper_bgcolor="rgba(256,256,256,1)",
            plot_bgcolor="rgba(256,256,256,0)",
        ),
        palettes=Palettes(),
        piemap=PieMapStyle(),
        tree=TreeStyle(),
        legend=LegendStyle(),
        show_legend=True,
        row_col_numbers=True,
        xstep=1,
        ystep=1,
    )

    def __init__(
        self,
        boxsize: Optional[int] = None,
        config: Optional[Dict[str, Any]] = None,
        highlight_annotation: Optional[str] = None,
        highlight_color: Optional[str] = None,
        labels: Optional[Union[LabelsStyle, Dict[str, Any]]] = None,
        layout: Optional[Union[go.Layout, Dict[str, Any]]] = None,
        palettes: Optional[Palettes] = None,
        piemap: Optional[Union[PieMapStyle, Dict[str, Any]]] = None,
        tree: Optional[Union[TreeStyle, Dict[str, Any]]] = None,
        legend: Optional[Union[LegendStyle, Dict[str, Any]]] = None,
        show_legend: Optional[bool] = None,
        row_col_numbers: Optional[bool] = None,
        xstep: Optional[str] = None,
        ystep: Optional[str] = None,
    ):
        super().__init__(
            boxsize=boxsize,
            config=config,
            highlight_annotation=highlight_annotation,
            highlight_color=highlight_color,
            labels=labels,
            layout=layout,
            palettes=palettes,
            piemap=piemap,
            tree=tree,
            legend=legend,
            show_legend=show_legend,
            row_col_numbers=row_col_numbers,
            xstep=xstep,
            ystep=ystep,
        )

    @field
    def boxsize(self) -> int:  # type: ignore
        """Size in pixels per row or column."""

    @field
    def config(self) -> Dict[str, Any]:  # type: ignore
        """Plotly Figure.show() config."""

    @field
    def highlight_annotation(self) -> Optional[str]:  # type: ignore
        """Annotation key of instances that should be highlighted.
        Value should be True-ish. Set key to `None` to disable."""

    @field
    def highlight_color(self) -> str:  # type: ignore
        """Default color to use for highlights."""

    @field
    def labels(self) -> LabelsStyle:  # type: ignore
        """Labels plot style."""

    @field
    def layout(self) -> go.Layout:  # type: ignore
        """Layout options."""

    @field
    def palettes(self) -> Palettes:  # type: ignore
        """Plot palettes options."""

    @field
    def piemap(self) -> PieMapStyle:  # type: ignore
        """Piechart map plot style."""

    @field
    def tree(self) -> TreeStyle:  # type: ignore
        """Tree plot style."""

    @field
    def legend(self) -> LegendStyle:  # type: ignore
        """Legend plot style."""

    @field
    def show_legend(self) -> bool:  # type: ignore
        """Boolean to display a legend."""

    @field
    def row_col_numbers(self) -> bool:  # type: ignore
        """Boolean to display row and column numbers."""

    @field
    def xstep(self) -> float:  # type: ignore
        """Axis increment per row or column in plots (usually 1)."""

    @field
    def ystep(self) -> float:  # type: ignore
        """Axis increment per row or column in plots (usually 1)."""


class Component(Mapping):
    """Plot component. The basic building block to create compound Plotly figures with.

    Arguments:
        width: Width in pixels.
        height: Height in pixels.
        traces: Traces to plot in this domain.
        shapes: SVG shapes from this component.
        xaxis: Plotly X-axis options.
        yaxis: Plotly Y-axis options.
    """

    _defaults: Dict[str, Any] = dict(
        width=0.0,
        height=0.0,
        traces=[],
        shapes=[],
        annotations=[],
        xaxis=go.layout.XAxis(),
        yaxis=go.layout.YAxis(),
    )

    def __init__(
        self,
        width: Optional[float] = None,
        height: Optional[float] = None,
        traces: Optional[List[BaseTraceType]] = None,
        shapes: Optional[List[Dict[str, Any]]] = None,
        annotations: Optional[List[Dict[str, Any]]] = None,
        xaxis: Optional[Union[go.layout.XAxis, Dict[str, Any]]] = None,
        yaxis: Optional[Union[go.layout.YAxis, Dict[str, Any]]] = None,
    ):
        if isinstance(xaxis, dict):
            _xaxis = deepcopy(self._defaults["xaxis"])
            for k, v in xaxis.items():
                setattr(_xaxis, k, v)
            xaxis = _xaxis

        if isinstance(yaxis, dict):
            _yaxis = deepcopy(self._defaults["yaxis"])
            for k, v in yaxis.items():
                setattr(_yaxis, k, v)
            yaxis = _yaxis

        super().__init__(
            width=width,
            height=height,
            traces=traces,
            shapes=shapes,
            annotations=annotations,
            xaxis=xaxis,
            yaxis=yaxis,
        )

    @field
    def width(self) -> float:  # type: ignore
        """Width in pixels."""

    @field
    def height(self) -> float:  # type: ignore
        """Height in pixels."""

    @field
    def traces(self) -> List[BaseTraceType]:  # type: ignore
        """Traces to plot in this domain."""

    @field
    def shapes(self) -> List[Dict[str, Any]]:  # type: ignore
        """SVG shapes from this component."""

    @field
    def annotations(self) -> List[Dict[str, Any]]:  # type: ignore
        """Annotations from this component."""

    @field
    def xaxis(self) -> go.layout.XAxis:  # type: ignore
        """Plotly X-axis options."""

    @field
    def yaxis(self) -> go.layout.YAxis:  # type: ignore
        """Plotly Y-axis options."""

    def get_figure(self, style: Style = Style(), show: bool = True) -> Optional[go.Figure]:
        """Get a Plotly figure of this component alone."""
        from ragraph.plot.utils import get_subplots, process_fig

        fig = get_subplots([[self]], style=style)
        return process_fig(fig, style=style, show=show)



"""# Donald B. Johnson's nested circuit finding algorithm"""
from collections import defaultdict
from typing import Dict, Generator, List, Optional, Set, Union

from ragraph.analysis.cluster._tarjan import tarjans_scc_algorithm
from ragraph.graph import Graph
from ragraph.node import Node


def johnson(
    graph: Graph,
    nodes: Optional[List[Node]] = None,
    names: bool = False,
    inherit: bool = True,
    **kwargs,
) -> Generator[Union[List[str], List[Node]], None, None]:
    """Donald B. Johnson's nested circuit finding algorithm. A circuit is a cycle in a
    graph, circuits can overlap or contain duplicate parts.

    Arguments:
        graph: Graph to find circuits in.
        nodes: Nodes to consider during circuit search.
        inherit: Whether to take into account (inherit) edges between children during
            SCC calculations.

    Yields:
        Node circuit lists.

    Note:
        Cannot cluster a graph directly since circuits may overlap in all sorts of ways.
        Therefore, it gives you cluster-related information, but no clear hierarchy.

    Reference:
        Johnson, D. B. (1975). Finding all the elementary circuits of a directed graph.
            In SIAM J. COMPUT (Vol. 4). https://doi.org/10.1137/0204007
    """
    if not nodes:
        nodes = graph.leafs

    # Keep original nodes around for node remapping, get a working copy for the algo.
    translation = {n.name: n for n in nodes}
    graph = graph.get_graph_slice(nodes, inherit=inherit)

    def _unblock(node: Node, blocked: Set[Node], B: Dict[Node, Set[Node]]):
        stack = set([node])
        while stack:
            node = stack.pop()
            if node in blocked:
                blocked.remove(node)
                stack.update(B[node])
                B[node].clear()

    # List of Lists of strongly connected nodes
    sccs = tarjans_scc_algorithm(graph, nodes=graph.roots, inherit=inherit)
    while sccs:
        scc = sccs.pop()
        startnode = scc.pop()
        path = [startnode]
        blocked = set()
        closed = set()
        blocked.add(startnode)
        B: Dict[Node, Set[Node]] = defaultdict(set)
        stack = [(startnode, list(graph.targets_of(startnode)))]
        while stack:
            node, targets = stack[-1]
            if targets:
                target = targets.pop()
                if target == startnode:
                    if names:
                        yield [translation[n.name].name for n in path]  # Original names
                    else:
                        yield [translation[n.name] for n in path]  # Original nodes
                    closed.update(path)
                elif target not in blocked:
                    path.append(target)
                    stack.append((target, list(graph.targets_of(target))))
                    closed.discard(target)
                    blocked.add(target)
                    continue
            else:
                if node in closed:
                    _unblock(node, blocked, B)
                else:
                    for target in graph.targets_of(node):
                        if node not in B[target]:
                            B[target].add(node)
                stack.pop()
                path.pop()
        graph.del_node(node)
        sccs.extend(tarjans_scc_algorithm(graph, scc, inherit))



"""Sequence nodes by name."""
from typing import List, Optional, Tuple, Union

from ragraph.analysis._classes import SequenceAnalysis
from ragraph.graph import Graph, Node

sort_by_name_analysis = SequenceAnalysis("Sort by name")


@sort_by_name_analysis
def name(
    graph: Graph,
    root: Optional[Union[str, Node]] = None,
    nodes: Optional[Union[List[str], List[Node]]] = None,
    inherit: bool = True,
    loops: bool = False,
    edge_weights: Optional[List[str]] = None,
    inplace: bool = True,
    names: bool = False,
    safe: bool = True,
    **kwargs,
) -> Tuple[Graph, List[Node]]:
    """Sequence nodes by node name.

    Arguments:
        graph: Graph to sequence nodes of.
        nodes: Nodes to sequence.

    Returns:
        Sequenced nodes.
    """
    return graph, sorted(nodes, key=lambda n: n.name)



"""# Kodak Single-Use Camera product commonality

This DSM dataset was actually proposed as a "3D DSM" as it was featured in the reference
papers in a 3D layered fashion. Relationships between components in a product family for
different family members were shown in a 3D layered fashion. Common interfaces and
unique interfaces were color coded accordingly for three family members: The "Fun
Saver", "Outdoor", and "Plus Digital" single-use cameras.

This dataset contains all unique product family components as nodes, where each node is
labeled with the family members it is used in. E.g. the node labels can be any
combination of at least one of the camera type.

Edges indicate an interface between components, which is therefore modeled using an
"adjacency" edge weight. Once again, they are labeled with the camera typename when that
dependency occurs in the corresponding camera type.

Available labels for nodes or edges:

- Fun Saver
- Outdoor
- Plus Digital

Reference:
    Alizon, Fabrice. 2009, February. Module-Based Design Management-Synerg'. Symposium on
    Product Family & Product Platform Design, Helsinki University of Technology (TKK),
    Helsinki, Finland.

Reference:
    Alizon, Fabrice, Seung K. Moon, Steven B. Shooter, and Timothy W. Simpson. 2007,
    September 4--7. Three Dimensional Design Structure Matrix-DSM3D. ASME Design Engineering
    Technical Conferences, DETCZ007-34510, Las Vegas, NV, pp. 941-948.
"""


edge_weights = ["adjacency"]



"""# Piechart mapping matrix plot component

The mapping plot is a matrix of pie-charts, indicating the edges between leaf nodes.

The nodes are identically placed from left-to-right as well as top-to-bottom. This
results in a diagonal (from top-left to bottom-right) of 'self loops', which are
commonly used to indicate "the nodes themselves".

The pie-charts can represent a number of things, including edge weights, edge labels as
well as other metrics. The node hierarchy is included using squares drawn around the
diagonal.

Furthermore, bus nodes have their respective row and column in a shaded background
color. This sets apart the buses' "highly integrative" edges from the regular edges
between nodes.
"""
from collections import OrderedDict, defaultdict
from copy import deepcopy
from math import cos, pi, prod, sin
from typing import Any, Dict, List, Tuple

import plotly.graph_objs as go

from ragraph import colors
from ragraph.edge import Edge
from ragraph.node import Node
from ragraph.plot import svg
from ragraph.plot.generic import Component, Style


class PieMap(Component):
    """A map of piecharts plot component.

    This map, or matrix, of pie-charts display the edges between leaf nodes.

    Arguments:
        rows: The nodes to be placed on the rows of the matrix.
        cols: The columns to be placed on the columns of the matrix.
        edges: Edges to be displayed between leaf nodes.
        style: Plot style option mapping.

    Note:
        The pie-charts can represent a number of things, including edge weights, edge
        labels as well as other metrics. The node hierarchy is included using squares
        drawn around the diagonal.

        Furthermore, bus nodes have their respective row and column in a shaded
        background color. This sets apart the buses' "highly integrative" edges from the
        regular edges between nodes.

    """

    def __init__(
        self,
        rows: List[Node],
        cols: List[Node],
        edges: List[Edge] = [],
        style: Style = Style(),
        **kwargs,
    ):
        # Custom grid.
        grid_shapes = _get_grid_shapes(rows, cols, style)

        # Display data
        pie_traces, pie_shapes = _get_pies_data(rows, cols, edges, style)

        hierarchy_shapes, kind_shapes = [], []
        if rows == cols:
            # Matrix is symmetric.
            # Shapes to indicate hierarchy.
            hierarchy_shapes = _get_hierarchy_shapes(rows, style)

            # Divide different node kinds by special lines.
            kind_shapes = _get_kind_shapes(rows, style)

        highlight_shapes = _get_highlight_shapes(rows, cols, style)

        # Combine all traces and shapes.
        traces = pie_traces
        shapes = highlight_shapes + grid_shapes + pie_shapes + kind_shapes + hierarchy_shapes

        dim = (len(rows), len(cols))
        xaxis, yaxis = deepcopy(style.piemap.xaxis), deepcopy(style.piemap.yaxis)
        xaxis.update(range=[0, dim[1]], scaleanchor="y", scaleratio=1.0)
        yaxis.update(range=[0, dim[0]])

        super().__init__(
            width=dim[1] * style.boxsize,
            height=dim[0] * style.boxsize,
            traces=traces,
            shapes=shapes,
            xaxis=xaxis,
            yaxis=yaxis,
        )


def _get_grid_shapes(rows: List[Node], cols: List[Node], style: Style):
    """Get grid traces and shapes for the mapping plot component.

    Arguments:
        rows: The nodes to be placed on the rows of the matrix.
        cols: The columns to be placed on the columns of the matrix.
        style: Plot style mapping.
    """
    dim = (len(rows), len(cols))

    line = style.piemap.gridline
    xstep, ystep = style.xstep, style.ystep

    # Gridlines
    shapes = [
        svg.get_line(x0=0, x1=dim[1] * xstep, y0=i * ystep, y1=i * ystep, line=line).as_dict()
        for i in range(dim[0] + 1)
    ] + [
        svg.get_line(x0=j * xstep, x1=j * xstep, y0=0, y1=dim[0] * ystep, line=line).as_dict()
        for j in range(dim[1] + 1)
    ]

    if rows == cols:
        shapes += [
            svg.get_rectangle(
                x0=i,
                x1=i + 1,
                y0=dim[0] - 1 - i,
                y1=dim[0] - i,
                fillcolor=line.color,
                line=dict(color=line.color),
                layer="below",
            ).as_dict()
            for i in range(dim[1])
        ]

    return shapes


def _get_highlight_shapes(rows: List[Node], cols: List[Node], style: Style) -> List[Dict[str, Any]]:
    """Get highlight bands for both rows and columns."""
    row_annot = (
        style.piemap.highlight_row_annotation
        if style.piemap.highlight_row_annotation
        else style.highlight_annotation
    )
    col_annot = (
        style.piemap.highlight_col_annotation
        if style.piemap.highlight_col_annotation
        else style.highlight_annotation
    )
    dim = (len(rows), len(cols))
    xstep, ystep = style.xstep, style.ystep

    shapes = []

    if col_annot:
        for i, node in enumerate(cols):
            value = node.annotations.get(col_annot)
            if not value:
                continue
            if isinstance(value, str):
                color = value
            else:
                color = (
                    style.piemap.highlight_col_color
                    if style.piemap.highlight_col_color
                    else style.highlight_color
                )
            shapes.append(
                svg.get_rectangle(
                    i * xstep,
                    (i + 1) * xstep,
                    0,
                    dim[0] * ystep,
                    fillcolor=color,
                    line=svg.Line(width=0),
                    layer="below",
                ).as_dict()
            )

    if row_annot:
        for i, node in enumerate(rows):
            value = node.annotations.get(row_annot)
            if not value:
                continue
            if isinstance(value, str):
                color = value
            else:
                color = (
                    style.piemap.highlight_row_color
                    if style.piemap.highlight_row_color
                    else style.highlight_color
                )
            shapes.append(
                svg.get_rectangle(
                    0,
                    dim[1] * xstep,
                    (dim[0] - 1 - i) * ystep,
                    (dim[0] - i) * ystep,
                    fillcolor=color,
                    line=svg.Line(width=0),
                    layer="below",
                ).as_dict()
            )

    return shapes


def _get_pies_data(rows: List[Node], cols: List[Node], edges: List[Edge], style: Style):
    """Get piechart traces and shapes for the mapping plot component.

    Arguments:
        rows: List of row nodes.
        cols: List of col nodes.
        edges: List of edges to draw data from.
        style: Plot style mapping.
    """
    # Dimensions
    dim = (len(rows), len(cols))

    # Get fields from style object or select all for the set display setting.
    fields = _get_fields(edges, style)

    # Check if categorical color settings are correct.
    _check_categorical_color_settings(fields=fields, style=style)

    # Prepare equal wedge angles once if that's the mode.
    equal_angles = _get_equal_angles(fields) if style.piemap.mode == "equal" else None

    # Get edge bundles.
    bundle_dict = _get_bundle_dict(rows, cols, edges, style)

    # Get fields dict.
    fields_dict, field_lower, field_upper = _get_fields_dict(bundle_dict, fields, style)

    # Calculate actual traces and shapes.
    traces = []
    shapes = []
    for i, source in enumerate(cols):
        for j, target in enumerate(rows):
            bundle = bundle_dict[source.name, target.name]
            if not bundle:
                continue  # No edges here.
            field_values = fields_dict[source.name, target.name]

            # Equal angles if calculated else calculate relative ones.
            field_angles = (
                _get_relative_angles(fields, field_values) if equal_angles is None else equal_angles
            )

            # Get the scaled radius.
            radius = (
                style.piemap.radius
                if style.piemap.scale_weight is None
                else style.piemap.radius
                * prod(e.weights.get(style.piemap.scale_weight, 1.0) for e in bundle)
            )

            # Draw the pie chart.
            customdata = _get_customdata(bundle, style)
            customdata.update(dict(source=source.name, target=target.name))
            pie_traces, pie_shapes = _get_pie_data(
                fields,
                field_angles,
                field_values,
                field_lower,
                field_upper,
                style,
                customdata=customdata,
                x=i + 0.5,
                y=dim[0] - j - 0.5,
                r=radius,
            )
            traces.extend(pie_traces)
            shapes.extend(pie_shapes)

    return traces, shapes


def _get_bundle_dict(
    rows: List[Node], cols: List[Node], edges: List[Edge], style: Style
) -> Dict[Tuple[str, str], List[Edge]]:
    """Get (inherited) edge lists between sources (cols) and targets (rows)."""
    # Basic edge mapping.
    edge_dict: Dict[Tuple[str, str], List[Edge]] = defaultdict(list)
    for e in edges:
        edge_dict[e.source.name, e.target.name].append(e)

    # Without inheritance, this edge dict is just the result we want.
    if not style.piemap.inherit:
        return edge_dict

    # Calculate edge bundles:
    bundle_dict: Dict[Tuple[str, str], List[Edge]] = defaultdict(list)
    for i, source in enumerate(cols):
        for j, target in enumerate(rows):
            bundle = edge_dict[source.name, target.name]
            bundle.extend(
                e
                for s in source.descendant_gen
                for t in target.descendant_gen
                for e in edge_dict[s.name, t.name]
            )
            bundle_dict[source.name, target.name] = bundle

    return bundle_dict


def _get_fields_dict(
    bundle_dict: Dict[Tuple[str, str], List[Edge]], fields: List[str], style: Style
) -> Tuple[Dict[Tuple[str, str], Dict[str, float]], Dict[str, float], Dict[str, float]]:
    """Get field values (display values) for each source/target combination."""
    # Get all field values.
    # Calculate once re-use later.
    fields_dict: Dict[Tuple[str, str], Dict[str, float]] = defaultdict(dict)
    field_lower: Dict[str, float] = dict()
    field_upper: Dict[str, float] = dict()
    for (src, tgt), bundle in bundle_dict.items():
        # Get values for the selected fields in this display mode.
        field_values = _get_field_values(bundle, fields, style)
        fields_dict[src, tgt] = field_values

        # Update lower/upper bounds.
        for field, val in field_values.items():
            if field not in field_lower or val < field_lower[field]:
                field_lower[field] = val
            if field not in field_upper or val > field_upper[field]:
                field_upper[field] = val
    return fields_dict, field_lower, field_upper


def _get_customdata(edges: List[Edge], style: Style) -> Dict[str, Any]:
    """Get customdata from edges."""
    customdata = {}

    for key in style.piemap.customhoverkeys:
        data = []
        for e in edges:
            if not hasattr(e.annotations, key):
                continue
            data.append(str(e.annotations[key]))

        customdata[key] = "<br>".join(data)

    return customdata


def _get_fields(edges: List[Edge], style: Style) -> List[str]:
    """Get fields from style mapping or default to all available fields to display."""
    style_fields = style.piemap.fields
    if style_fields is not None:
        return style_fields

    display = style.piemap.display
    if display == "kinds":
        fields = set(e.kind for e in edges)
    elif display == "labels":
        fields = set(label for e in edges for label in e.labels)
    elif display in {"weight labels", "weights"}:
        fields = set(wl for e in edges for wl in e.weights.keys())
    else:
        fields = set()

    return sorted(fields)


def _check_categorical_color_settings(fields: List[str], style: Style):
    """Check if categorical colors list is of sufficient length (len=10 by default)."""
    if style.piemap.display == "weights":
        return

    if len(fields) > len(style.palettes.categorical):
        style.palettes.categorical = colors.get_categorical(len(fields))


def _get_field_values(edges: List[Edge], fields: List[str], style: Style) -> Dict[str, float]:
    """Get the field sums for the given edges according to the style."""
    field_map: Dict[str, float] = dict()
    if not edges:
        return field_map

    display = style.piemap.display
    if display == "kinds":
        field_map = {
            f: float(sum(f == e.kind for e in edges))
            for f in fields
            if sum(f == e.kind for e in edges)
        }
    elif display == "labels":
        field_map = {
            f: float(sum(f in e.labels for e in edges))
            for f in fields
            if sum(f in e.labels for e in edges)
        }
    elif display == "weight labels":
        field_map = {
            f: float(sum(f in e.weights for e in edges))
            for f in fields
            if sum(f in e.weights for e in edges)
        }
    elif display == "weights":
        field_map = {f: float(sum(e.weights.get(f, 0.0) for e in edges)) for f in fields}
    else:
        field_map = dict(num=float(len(edges)))
    return field_map


def _get_equal_angles(fields: List[str]) -> Dict[str, Tuple[float, float]]:
    """Get piechart angle dictionary."""
    delta = 2 * pi / len(fields)
    offset = pi / 2  # Start upright, then go clockwise for intuitive results.
    return {f: (offset - (i + 1) * delta, offset - i * delta) for i, f in enumerate(fields)}


def _get_relative_angles(
    fields: List[str], sums: Dict[str, float]
) -> Dict[str, Tuple[float, float]]:
    """Get the field pie-chart angles given their value sums."""
    total = sum(abs(v) for v in sums.values())
    angles = dict()
    if not total:
        return _get_equal_angles(fields)

    end = pi / 2
    for field in fields:
        if not sums.get(field):
            continue

        delta = 2 * pi * abs(sums[field]) / total
        start = end - delta
        angles[field] = (start, end)
        end = start

    return angles


def _get_pie_data(
    fields: List[str],
    angles: Dict[str, Tuple[float, float]],
    values: Dict[str, float],
    lower: Dict[str, float],
    upper: Dict[str, float],
    style: Style,
    customdata: Dict[str, Any],
    x: float = 0.0,
    y: float = 0.0,
    r: float = 0.4,
):
    """Draw pie charts in the order of the given fields."""
    display = style.piemap.display

    shapes = []
    xdata = []
    ydata = []
    textdata = []

    # Draw for every field.
    for i, field in enumerate(fields):
        if field not in angles or field not in values:
            continue

        # Get the right color.
        if display == "weights":
            color = style.palettes.get_continuous_color(
                values[field], lower[field], upper[field], field=field
            )
        else:
            color = style.palettes.get_categorical_color(i, field=field)

        # Fetch angles.
        start, end = angles[field]

        # Append shape.
        shapes.append(
            svg.get_wedge(
                x=x,
                y=y,
                r=r,
                start_angle=start,
                end_angle=end,
                fillcolor=color,
                line=dict(width=0),
            ).as_dict(),
        )

        # Calculate and append trace data.
        delta = end - start
        xdata.append(x + 0.5 * r * cos(start + 0.5 * delta))
        ydata.append(y + 0.5 * r * sin(start + 0.5 * delta))

        text = (
            f"source: {customdata['source']}"
            + f"<br>target: {customdata['target']}"
            + f"<br>{field}: {values[field]}"
        )

        for key in style.piemap.customhoverkeys:
            text += "<br>" + f"{key}: {customdata[key]}"

        textdata.append(text)

    # Get traces.
    traces = [
        go.Scatter(
            x=xdata,
            y=ydata,
            text=textdata,
            mode="markers",
            marker=dict(color="#ffffff"),
            hoverinfo="text",
            showlegend=False,
            customdata=[customdata],
        )
    ]

    return traces, shapes


def _get_hierarchy_shapes(leafs: List[Node], style: Style) -> List[Dict[str, Any]]:
    """Get hierarchy shapes for the mapping plot component.

    Arguments:
      leafs: List of leaf nodes.
      style: Style of the tree to be plotted.

    Returns:
        Hierarchy shapes as SVG mappings converted to dictionaries.
    """

    shapes: List[Dict[str, Any]] = []

    # Plot geometrics
    x_step, y_step = style.xstep, style.ystep
    tree_depth = max([node.depth for node in leafs])
    dim = len(leafs)

    # Compute leaf node positions.
    odict = OrderedDict()
    sdict = OrderedDict()
    for i, n in enumerate(leafs):
        odict[n] = i  # Offset dict
        sdict[n] = 1  # Size dict

    # Compute ancestor positions.
    processed_ancestors = set()
    parents = set(node.parent for node in leafs if node.parent)
    current_depth = tree_depth - 1
    while parents:
        grandparents = set()
        for parent in parents:
            if parent in processed_ancestors:
                continue
            if parent.depth < current_depth:
                grandparents.add(parent)
                continue

            # Only take into account children that are visible.
            children = [child for child in parent.children if child in odict]

            # Offset is minimum value.
            offset = min(odict[n] for n in children)
            odict[parent] = offset

            # Size is sum of child sizes.
            size = sum(sdict[n] for n in children)
            sdict[parent] = size

            shapes.append(
                svg.get_rectangle(
                    x0=offset * x_step,
                    x1=(offset + size) * x_step,
                    y0=(dim - offset - size) * y_step,
                    y1=(dim - offset) * y_step,
                    line=style.piemap.clusterline,
                ).as_dict()
            )

            processed_ancestors.add(parent)
            if parent.parent:
                grandparents.add(parent.parent)

        parents = grandparents
        current_depth -= 1

    busarea = style.piemap.busarea

    for node in sdict:
        if node.is_bus and node.parent in sdict:
            assert node.parent is not None
            parent_size = sdict[node.parent]
            parent_offset = odict[node.parent]
            node_offset = odict[node]
            node_size = sdict[node]

            # Horizontal bus bar.
            shapes.append(
                svg.get_rectangle(
                    x0=parent_offset * x_step,
                    x1=(parent_offset + parent_size) * x_step,
                    y0=(dim - node_offset - node_size) * y_step,
                    y1=(dim - node_offset) * y_step,
                    **busarea.as_dict(),
                ).as_dict()
            )

            # Vertical bus bar.
            shapes.append(
                svg.get_rectangle(
                    x0=node_offset * x_step,
                    x1=(node_offset + node_size) * x_step,
                    y0=(dim - parent_offset - parent_size) * y_step,
                    y1=(dim - parent_offset) * y_step,
                    **busarea.as_dict(),
                ).as_dict()
            )

    return shapes


def _get_kind_shapes(leafs: List[Node], style: Style):
    """Get kind shapes for the mapping plot component."""
    dim = len(leafs)
    line = style.piemap.kindline

    # Get the indexes of leafs that "change" the kind and append lines afterwards.
    kind = leafs[0].kind
    shapes = []
    for i, node in enumerate(leafs):
        if node.kind == kind:
            continue
        kind = node.kind

        # Calculate y coordinate of horizontal line.
        y = (dim - i) * style.ystep

        # Horizontal line.
        shapes.append(svg.get_line(x0=0, x1=dim, y0=y, y1=y, line=line).as_dict())

        # Vertical line.
        shapes.append(svg.get_line(x0=i, x1=i, y0=0, y1=dim, line=line).as_dict())

    return shapes



"""# Hierarchy tree plot component

This module contains the [`Tree` component][ragraph.plot.components.tree.Tree] which produces a
[`Component`][ragraph.plot.generic.Component] for the hierarchy tree of a vertical list of leaf
nodes up to their roots which are put on their left.
"""
from collections import OrderedDict
from copy import deepcopy
from typing import Any, Dict, List, Set, Tuple, Union

import plotly.graph_objs as go

from ragraph.node import Node
from ragraph.plot import svg
from ragraph.plot.generic import Component, Style


class Tree(Component):
    """Hierarchy tree plot component of leaf nodes up to their roots.

    Leaf nodes are plotted as a vertical list with their roots on the left side.

    Arguments:
        leafs: List of leaf nodes.
        style: Plot style mapping.
    """

    def __init__(self, leafs: List[Node], style: Style = Style()):
        ancestors, depth, xdict, ydict = _get_data(leafs=leafs, style=style)

        node_trace = _get_node_trace(
            xdict=xdict,
            ydict=ydict,
            style=style,
        )
        line_shapes = _get_line_shapes(ancestors=ancestors, xdict=xdict, ydict=ydict, style=style)

        # Calculating geometric boundaries.
        width = (depth + 1) * style.xstep * style.boxsize
        height = len(leafs) * style.ystep * style.boxsize

        xaxis, yaxis = deepcopy(style.tree.xaxis), deepcopy(style.tree.yaxis)
        xaxis.update(range=(-0.5, width / style.boxsize - 0.5), scaleanchor="y", scaleratio=1.0)
        yaxis.update(range=(0, height / style.boxsize))

        super().__init__(
            width=width,
            height=height,
            traces=[node_trace],
            shapes=line_shapes,
            xaxis=xaxis,
            yaxis=yaxis,
        )


def _get_data(
    leafs: List[Node], style: Style
) -> Tuple[Set[Node], int, Dict[Node, float], Dict[Node, float]]:
    """Compute coordinates of tree nodes.

    Arguments:
      leafs: List of leaf nodes.
      style: Plot style mapping.

    Returns:
        Processed ancestor nodes.
        Tree depth.
        Dictionary of node to x position of tree nodes.
        Dictionary of node to y position of tree nodes.
    """

    # Plot geometrics
    x_step = style.xstep
    y_step = style.ystep
    tree_depth = max([node.depth for node in leafs])
    tree_height = len(leafs) * y_step

    # Compute leaf node positions.
    xdict = OrderedDict()
    ydict = OrderedDict()
    for idx, leaf in enumerate(leafs):
        xdict[leaf] = tree_depth * x_step
        ydict[leaf] = tree_height - (idx + 0.5) * y_step

    # Compute ancestor positions.
    processed_ancestors = set()
    parents = set(node.parent for node in leafs if node.parent)
    current_depth = tree_depth - 1
    while parents:
        grandparents = set()
        for parent in parents:
            if parent in processed_ancestors:
                continue
            if parent.depth < current_depth:
                grandparents.add(parent)
                continue

            # Calculate x positition based on parent depth.
            xdict[parent] = parent.depth * x_step

            # Calculate y position as the average of the min/max of child nodes.
            ys_children = [ydict[child] for child in parent.children if child in ydict]
            ymin = min(ys_children)
            ymax = max(ys_children)
            ydict[parent] = (ymin + ymax) / 2
            processed_ancestors.add(parent)

            if parent.parent:
                grandparents.add(parent.parent)

        parents = grandparents
        current_depth -= 1

    return processed_ancestors, tree_depth, xdict, ydict


def _get_node_trace(xdict: Dict[Node, float], ydict: Dict[Node, float], style: Style) -> go.Scatter:
    """Get tree node trace from positional dictionaries. These are the 'dots'.

    Arguments:
      xdict: Dictionary of node to x position of tree nodes.
      ydict: Dictionary of node to y position of tree nodes.
      style: Plot style mapping.

    Returns:
        Plotly Scatter trace.
    """
    xdata, ydata, textdata, customdata = zip(
        *[
            (xdict[n], ydict[n], n.name, {"pointtype": "tree_node", "data": n.name})
            for n in xdict.keys()
        ]
    )

    tree_node_trace = go.Scatter(
        x=xdata,
        y=ydata,
        text=textdata,
        mode="markers",
        marker={"color": style.tree.line.color},
        hoverinfo="text",
        showlegend=False,
        customdata=customdata,
    )

    return tree_node_trace


def _get_line_shapes(
    ancestors: Set[Node],
    xdict: Dict[Node, float],
    ydict: Dict[Node, float],
    style: Style,
) -> List[Dict[str, Any]]:
    """Compute the actual tree lines of the tree plot. These form the tree shapes.

    Arguments:
        ancestors: Set of processed ancestor nodes.
        xdict: Dictionary of node to x position of tree nodes.
        ydict: Dictionary of node to y position of tree nodes.
        style: Plot style mapping.

    Returns:
        List of SVG shapes.
    """
    svgs = []
    for node in ancestors:
        for child in node.children:
            if child not in xdict:
                continue
            svgs.extend(
                _get_parent_child_line(
                    x0=xdict[node],
                    y0=ydict[node],
                    x1=xdict[child],
                    y1=ydict[child],
                    style=style,
                )
            )
    shapes = [s.as_dict() for s in svgs]
    return shapes


def _get_parent_child_line(
    x0: Union[int, float],
    y0: Union[int, float],
    x1: Union[int, float],
    y1: Union[int, float],
    style: Style,
) -> List[svg.SVG]:
    """Draw a line with a curved 90 degree angle from a parent to a child in the tree.

    Arguments:
        x0: x starting coordinate of the line.
        y0: y starting coordinate of the line.
        x1: x ending coordinate of the line.
        y1: y ending coordinate of the line.
        style: Plot style mapping.

    Returns:
       List of SVG shape mappings of the different line sections.
    """
    linestyle = style.tree.line
    curve_delta = min(0.1, abs(y0 - y1))
    if y0 == y1:
        # Parent and child are aligned so only draw a straight line.
        return [svg.get_line(x0=x0, x1=x1, y0=y0, y1=y1, line=linestyle)]
    elif y0 > y1:
        # Draw line downwards.
        y = y1 + curve_delta
    elif y0 < y1:
        y = y1 - curve_delta

    x = x0 + curve_delta

    first_line = svg.get_line(x0=x0, x1=x0, y0=y0, y1=y, line=linestyle)
    curve_line = svg.get_curvedline(x0=x0, x1=x0, x2=x, y0=y, y1=y1, y2=y1, line=linestyle)
    second_line = svg.get_line(x0=x, x1=x1, y0=y1, y1=y1, line=linestyle)

    return [first_line, curve_line, second_line]



"""# Minimal compatibility analysis example

Contains 6 component variant nodes. They are divided in three node kinds (e.g. components), which
correspond to the first character in their node names: A1, B1, B2, C1, C2, C3. For ease of usage,
the "performance" weight of each node is set to it's node name's second character.

Compatibility between nodes is signalled using edges with a "compatibility" kind.
"""

node_weights = ["performance"]
